"""
Mixin classes related to vessel properties.

The following classes are provided:
- HasLength
- HasLoad
- Vesselproperties

"""
# package(s) for documentation, debugging, saving and loading
import logging

# math packages
import random

# spatial libraries
import networkx as nx

# package(s) related to the simulation
import simpy

# use OpenCLSim objects for core objects (identifiable is imported for later use)
from openclsim.core import SimpyObject
import opentnsim.graph.mixins

# get logger
logger = logging.getLogger(__name__)


class HasLength(SimpyObject):
    """Mixin class: Something with a length. The length is modelled as a storage capacity

    Parameters
    -----------
    length: float
        length that can be requested
    remaining_length: float, default=0
        length that is still available at the beginning of the simulation.
    args, kwargs:
        passed to SimpyObject. Must at least contain parameter env: simpy.Environment.

    Attributes
    -----------
    length: simpy.Container
        the container that is used to limit the length that can be requested.
    pos_length: simpy.Container
        the container that is used to limit the length that can be requested.
    """

    def __init__(self, length: float, remaining_length: float = 0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """Initialization"""
        self.length = simpy.Container(self.env, capacity=length, init=remaining_length)
        self.pos_length = simpy.Container(self.env, capacity=length, init=remaining_length)


class HasLoad:
    """Mixin class with load dependent height (H) and draught (T). The filling
    degree (filling_degree: fraction) will interpolate between empty and full
    height and draught.

    Parameters
    ----------
    H_e: float
        height of the vessel when empty
    H_f: float
        height of the vessel when full
    T_e: float
        draught of the vessel when empty
    T_f: float
        draught of the vessel when full
    filling_degree: float, default=0
        fraction of the vessel that is filled upon creation

    """

    def __init__(self, H_e, H_f, T_e, T_f, filling_degree=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.H_e = H_e
        self.H_f = H_f
        self.T_e = T_e
        self.T_f = T_f
        self.filling_degree = filling_degree

    @property
    def T(self):
        """Calculate current draught based on filling degree"""
        T = self.filling_degree * (self.T_f - self.T_e) + self.T_e
        return T

    @property
    def H(self):
        """Calculate current height based on filling degree"""
        return self.filling_degree * (self.H_f - self.H_e) + self.H_e


class VesselProperties:
    """Mixin class: Something that has vessel properties
    This mixin is updated to better accommodate the ConsumesEnergy mixin

    Parameters
    ----------
    type: str
        can contain info on vessel type (avv class, cemt_class or other)
    B: float
        vessel width
    L: float
        vessel length
    h_min: float, optional
        vessel minimum water depth.
        The wather-depth can also be extracted from the network edges if they have the property ['Info']['GeneralDepth']
    T: float, optional
        actual draught.
        If not given, it will be calculated based on the filling degree (use mixin HasLoad) or based on the payload and vessel type
    safety_margin: float, optional
      the water area above the waterway bed reserved to prevent ship grounding due to ship squatting during sailing
      the value of safety margin depends on waterway bed material and ship types. For tanker vessel with rocky bed the safety
      margin is recommended as 0.3 m based on Van Dorsser et al. The value setting for safety margin depends on the risk attitude
      of the ship captain and shipping companies.
    h_squat: boolean, optional
        the water depth considering ship squatting while the ship moving (if False, squat is disabled, if True, squat is calculated)
    payload: float, optional
        cargo load [ton], the actual draught can be determined by knowing payload based on van Dorsser et al's method.
        (https://www.researchgate.net/publication/344340126_The_effect_of_low_water_on_loading_capacity_of_inland_ships)
    vessel_type: str, optional
        vessel_type: vessel type can be selected from "Container","Dry_SH","Dry_DH","Barge","Tanker".
        ("Dry_SH" means dry bulk single hull, "Dry_DH" means dry bulk double hull),
        based on van Dorsser et al's paper.
        (https://www.researchgate.net/publication/344340126_The_effect_of_low_water_on_loading_capacity_of_inland_ships)
    renewable_fuel_mass: float, optional
        renewable fuel mass on board [kg]
    renewable_fuel_volume: float, optional
        renewable fuel volume on board [m3]
    renewable_fuel_required_space: float, optional
        renewable fuel required storage space (consider packaging factor) on board  [m3]
    """

    def __init__(
        self,
        type,
        B,
        L,
        h_min=None,
        T=None,
        safety_margin=None,
        h_squat=None,
        payload=None,
        vessel_type=None,
        renewable_fuel_mass=None,
        renewable_fuel_volume=None,
        renewable_fuel_required_space=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        """Initialization
        """
        self.type = type
        self.B = B
        self.L = L
        # hidden because these can also computed on the fly
        self._T = T
        self._h_min = h_min
        # alternative  options
        self.safety_margin = safety_margin
        self.h_squat = h_squat
        self.payload = payload
        self.vessel_type = vessel_type
        self.renewable_fuel_mass = renewable_fuel_mass
        self.renewable_fuel_volume = renewable_fuel_volume
        self.renewable_fuel_required_space = renewable_fuel_required_space

        # for lock module: #TODO: op een andere manier toevoegen?
        self.bound = "inbound"

    @property
    def T(self):
        """Compute the actual draught.
        This will default to using the draught passed by the constructor. If it is None it will try to find one in the super class.
        """
        if self._T is not None:
            # if we were passed a T value, use tha one
            T = self._T
        elif self.T_f is not None and self.T_e is not None:
            # base draught on filling degree
            T = self.filling_degree * (self.T_f - self.T_e) + self.T_e
        elif self.payload is not None and self.vessel_type is not None:
            T = opentnsim.strategy.Payload2T(
                self,
                Payload_strategy=self.payload,
                vessel_type=self.vessel_type,
                bounds=(0, 40),
            )  # this need to be tested
        # todo: for later possibly include Payload2T

        return T

    @property
    def h_min(self):
        """get the minimum water depth. if not given, h_min is the minimal water depth of the graph.."""
        if self._h_min is not None:
            h_min = self._h_min
        else:
            h_min = opentnsim.graph.mixins.get_minimum_depth(graph=self.graph, route=self.route)

        return h_min

    def get_route(
        self,
        origin,
        destination,
        graph=None,
        minWidth=None,
        minHeight=None,
        minDepth=None,
        randomSeed=4,
    ):
        """Calculate a path based on vessel restrictions
        restrictions are only applied if the graph has the attributes ['Width', 'Height', 'Depth']

        Parameters
        ----------
        origin: str
            ID of the starting node
        destination: str
            ID of the destination node
        graph: networkx.Graph, default = self.graph
            The graph to use for the pathfinding
        minWidth: float, default = 1.1 * self.B
            Minimum width of the path
        minHeight: float, default = 1.1 * self.H
            Minimum height of the path
        minDepth: float, default = 1.1 * self.T
            Minimum depth of the path
        randomSeed: int, default = 4
            Seed for the random number generator

        """

        graph = graph if graph else self.graph
        minWidth = minWidth if minWidth else 1.1 * self.B
        minHeight = minHeight if minHeight else 1.1 * self.H
        minDepth = minDepth if minDepth else 1.1 * self.T

        # Check if information on restrictions is added to the edges
        random.seed(randomSeed)
        edge = random.choice(list(graph.edges(data=True)))
        edge_attrs = list(edge[2].keys())

        # IMPROVE THIS TO CHECK ALL EDGES AND COMBINATIONS OF RESTRICTIONS

        if all(item in edge_attrs for item in ["Width", "Height", "Depth"]):
            edges = []
            nodes = []

            for edge in graph.edges(data=True):
                if edge[2]["Width"] >= minWidth and edge[2]["Height"] >= minHeight and edge[2]["Depth"] >= minDepth:
                    edges.append(edge)

                    nodes.append(graph.nodes[edge[0]])
                    nodes.append(graph.nodes[edge[1]])

            subGraph = graph.__class__()

            for node in nodes:
                subGraph.add_node(
                    node["name"],
                    name=node["name"],
                    geometry=node["geometry"],
                    position=(node["geometry"].x, node["geometry"].y),
                )

            for edge in edges:
                subGraph.add_edge(edge[0], edge[1], attr_dict=edge[2])

            try:
                return nx.dijkstra_path(subGraph, origin, destination)
                # return nx.bidirectional_dijkstra(subGraph, origin, destination)
            except nx.NetworkXNoPath:
                raise ValueError("No path was found with the given boundary conditions.")

        # If not, return shortest path
        else:
            return nx.dijkstra_path(graph, origin, destination)



# -*- coding: utf-8 -*-
from importlib.metadata import version, PackageNotFoundError

import opentnsim.core as core  # now contains mixin class consumes energy, also the modified vessel properties
import opentnsim.energy as energy
import opentnsim.graph.mixins as mixins
import opentnsim.lock as lock
import opentnsim.model as model
import opentnsim.plot as plot
import opentnsim.strategy as strategy


"""Top-level package for OpenTNSim."""

__author__ = """Mark van Koningsveld"""
__email__ = "M.vanKoningsveld@tudelft.nl"

try:
    __version__ = version("opentnsim")
except PackageNotFoundError:
    # Package is not installed yet (e.g., running tests from source without an install)
    # This value will be replaced by the build system when the package is installed
    __version__ = "uninstalled" 



"""Module with alogorithms necessary for the energy module"""

# %% IMPORT DEPENENDENCIES
# generic
import pathlib
import logging
import functools
import numpy as np
import pandas as pd
import scipy.optimize

# OpenTNSim
import opentnsim

# logging
logger = logging.getLogger(__name__)


def power2v(vessel, edge, upperbound):
    """Compute vessel velocity given an edge and power (P_tot_given)

    bounds is the limits where to look for a solution for the velocity [m/s]
    returns velocity [m/s]
    """

    assert isinstance(vessel, opentnsim.core.vessel_properties.VesselProperties), "vessel should be an instance of VesselProperties"
    assert vessel.C_B is not None, "C_B cannot be None"

    def seek_v_given_power(v, vessel, edge):
        """function to optimize"""
        # TODO: check it this needs to be made more general, now relies on ['Info'] to be present
        # water depth from the edge
        h_0 = edge["Info"]["GeneralDepth"]
        try:
            h_0 = vessel.calculate_h_squat(v, h_0)
        except AttributeError:
            # no squat available
            pass
        # TODO: consider precomputing a range v/h combinations for the ship before the simulation starts
        vessel.calculate_total_resistance(v, h_0)

        # compute total power given
        P_given = vessel.calculate_total_power_required(v=v, h_0=h_0)
        if isinstance(vessel.P_tot, complex):
            raise ValueError(f"P tot is complex: {vessel.P_tot}")

        # compute difference between power setting by captain (incl hotel) and power needed for velocity (incl hotel)
        diff = vessel.P_tot_given - P_given  # vessel.P_tot
        logger.debug(f"optimizing for v: {v}, P_tot_given: {vessel.P_tot_given}, P_tot {vessel.P_tot}, P_given {P_given}")

        return diff**2

    # fill in some of the parameters that we already know
    fun = functools.partial(seek_v_given_power, vessel=vessel, edge=edge)
    # lookup a minimum
    fit = scipy.optimize.minimize_scalar(fun, bounds=(0, upperbound), method="bounded", options=dict(xatol=0.0000001))

    # check if we found a minimum
    if not fit.success:
        raise ValueError(fit)
    logger.debug(f"fit: {fit}")

    return fit.x



# config.py
import numpy as np

import xarray as xr

class HydrodynamicDataManager:
    """
    Singleton class to manage hydrodynamic data.
    This class ensures that hydrodynamic data is loaded only once and can be accessed globally.
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(HydrodynamicDataManager, cls).__new__(cls)
            cls._instance.hydrodynamic_data = None
            cls._instance.hydrodynamic_times = None

        return cls._instance

    def _get_hydrodynamic_data_value(self, hydrodynamic_information_path, time, node, hydrodynamic_property):
        """Gets the value of a hydrodynamic property at a certain time and node

        Parameters
        ----------
        hydrodynamic_information_path : path
            the path to hydrodynamic data (if None, no hydrodynamic data is included in the simulation)
        time : np.datetime64
            the time
        node : str
            the node name in the graph
        hydrodynamic_property : str
            the hydrodynamic property: "Water level", "Current velocity", "Salinity" (if included in the hydrodynamic data)

        Returns
        -------
        value : float
            the value of a hydrodynamic property at the specified time and node
        """
        value = np.nan
        if hydrodynamic_information_path is None:
            return value

        # determine the time_index and station_inex
        time_index = self._get_time_index_of_hydrodynamic_data(hydrodynamic_information_path, time)
        station_index = self._get_station_index_of_hydrodynamic_data(hydrodynamic_information_path, node)

        # determine the property
        if isinstance(self.hydrodynamic_data, xr.Dataset):
            value = self.hydrodynamic_data[hydrodynamic_property][station_index][time_index].values.copy()
        else:
            value = self.hydrodynamic_data[hydrodynamic_property][station_index][time_index].copy()

        return value

    def _get_time_index_of_hydrodynamic_data(self, hydrodynamic_information_path, time):
        """Gets the time index in the hydrodynamic data closest to a time

        Parameters
        ----------
        env : Simpy.Environment
            the simulation environment (to access the hydrodynamic data).
            the time

        Returns
        -------
        time_index : int
            the time index of the hydrodynamic data closest to the time
        """
        time_index = 0
        if hydrodynamic_information_path is None:
            return time_index

        # determine the time_index
        if isinstance(self.hydrodynamic_data, xr.Dataset):
            time_index = np.absolute(self.hydrodynamic_times - time).argmin().values
        else:
            time_index = np.absolute(self.hydrodynamic_times - time).argmin()

        return time_index

    def _get_station_index_of_hydrodynamic_data(self, hydrodynamic_information_path, node):
        """Gets the node's station index in the hydrodynamic data

        Parameters
        ----------
        hydrodynamic_information_path : path
            the path to hydrodynamic data (if None, no hydrodynamic data is included in the simulation)
        node : str
            the node name in the graph

        Returns
        -------
        station_index : str
            the time index of the hydrodynamic data closest to the time
        """

        station_index = 0
        if hydrodynamic_information_path is None:
            return station_index

        if isinstance(self.hydrodynamic_data, xr.Dataset):
            station_index = np.where(np.array(list((self.hydrodynamic_data["STATION"].values))) == node)[0][0]
        else:
            station_index = np.where(np.array(list((self.hydrodynamic_data["STATION"]))) == node)[0]

        return station_index

    def _get_hydrodynamic_data_series(self, hydrodynamic_information_path, time, node, hydrodynamic_property):
        """Gets the time series of a hydrodynamic property at a certain node from a certain time onwards

        Parameters
        ----------
        hydrodynamic_information_path : path
            the path to hydrodynamic data (if None, no hydrodynamic data is included in the simulation)
        time : np.datetime64
            the time
        node : str
            the node name in the graph
        hydrodynamic_property : str
            the hydrodynamic property: "Water level", "Current velocity", "Salinity" (if included in the hydrodynamic data)

        Returns
        -------
        series : float
            the time series of a hydrodynamic property at the specified node from the specified time onwards
        """
        series = np.array([np.nan])
        if hydrodynamic_information_path is None:
            return series

        # determine the time_index and station_inex
        time_index = self._get_time_index_of_hydrodynamic_data(hydrodynamic_information_path, time)
        station_index = self._get_station_index_of_hydrodynamic_data(hydrodynamic_information_path, node)

        # determine the property
        if isinstance(self.hydrodynamic_data, xr.Dataset):
            series = self.hydrodynamic_data[hydrodynamic_property][station_index][time_index:].values.copy()
        else:
            series = self.hydrodynamic_data[hydrodynamic_property][station_index][time_index:].copy()

        return series



"""This module contains utility functions for lock operations in the OpenTNSim simulation environment."""

import logging

import pandas as pd

logger = logging.getLogger(__name__)


def _get_lock_operation_to_and_from_node(lock, direction):
    """Get the nodes from and to which the lock operation is directed based on the direction convention

    Convention: direction = 0 (when A -> B), direction = 1 (when B -> A)
        with A = lock.start_node and B = lock.end_node

    Parameters
    ----------
    lock : object
        the lock object generated with IsLockComplex
    direction : int
        the direction of the lock operation: 0 or 1 (see above convention)

    Returns
    -------
    node_of_approach : str
        the name of the node from which the lock operation is directed
    to_node : str
        the name of the node to which the lock operation is directed
    """
    node_of_approach = lock.end_node
    to_node = lock.start_node
    if not direction:
        node_of_approach = lock.start_node
        to_node = lock.end_node
    return node_of_approach, to_node


def _get_lock_object_on_registration_node(multidigraph, registration_node):
    """Get the lock complex object that is associated with a registration node node

    Parameters
    ----------
    multidigraph : nx.MultiDiGraph
        the graph of the simulation as MultiDiGraph-version (to allow for parallel locks between the same node pair)
    registration_node : str
        node name (that has to be in the graph) on which the vessel is currently starting to navigate an edge

    Returns
    -------
    lock : Union(class, None)
        the lock complex object that is associated with the registration node, or None if no lock complex is associated with the registration node
    """
    # check if node is a registration node
    if "Lock_registration_node" not in multidigraph.nodes[registration_node].keys():
        return None

    edge = multidigraph.nodes[registration_node]["Lock_registration_node"]
    # return lock if it exists on the edge
    if "Lock" in multidigraph.edges[edge].keys():
        lock = multidigraph.edges[edge]["Lock"][0]
        return lock
    # Return None if no lock exists on the edge
    else:
        return None


def _update_lock_operation_planning(lock, operation_index, operation_information):
    """Updates the lock operation planning

    Parameters
    ----------
    lock : object
        the lock object generated with IsLockComplex
    operation_index : int
            index of the lock operation
    operation_information : dict
        information to be added to the dataframe with keys as column names and values
    """
    for key, value in operation_information.items():
        if key not in lock.operation_planning.columns:
            warnings.warn(f"Column name ({key}) not in the operation planning dataframe -> skipped.")
            continue

        # explictly set 1 value (so we can set a list as a value)
        lock.operation_planning.at[int(operation_index), key] = value




def _update_lock_vessel_planning(lock, vessel_index, passage_information):
    """Updates the lock vessel planning

    Parameters
    ----------
    lock : object
        the lock object generated with IsLockComplex
    operation_index : int
            index of the lock operation
    passage_information : dict
        information to be added to the dataframe with keys as column names and values
    """
    for key, value in passage_information.items():
        if key not in lock.vessel_planning.columns:
            warnings.warn(f"Column name ({key}) not in the vessel planning dataframe -> skipped.")
            continue

        if isinstance(value, pd.Timedelta):
            # round to us
            value = value.round('us')

        lock.vessel_planning.at[int(vessel_index), key] = value
        

def determine_route_to_closest_waiting_area(vessel, waiting_area_A, waiting_area_B):
    """
    DOCUMENTATION HERE

    :param node:
    :param vessel:
    :return:
    """

    remaining_route = vessel.route_ahead
    waiting_area_node = None
    for origin in remaining_route:
        if origin == waiting_area_A.edge[0]:
            waiting_area_node = waiting_area_A.edge[1]
            break
        elif origin == waiting_area_B.edge[0]:
            waiting_area_node = waiting_area_B.edge[1]
            break
    if waiting_area_node is not None:
        route_to_waiting_area = vessel.determine_route_to_target_node(target_node=waiting_area_node)
    else:
        route_to_waiting_area = []
        warnings.warn(f"No route found to waiting area")
    return route_to_waiting_area


def _get_lock_operation_direction(lock, to_node):
    """Get the direction of the lock based on the node to which the lock operation is directed

    Convention: direction = 0 (when A -> B), direction = 1 (when B -> A)
        with A = lock.start_node and B = lock.end_node

    Parameters
    ----------
    lock : object
        the lock object generated with IsLockComplex
    to_node : str
        the name of the node to which the lock operation is directed

    Returns
    -------
    direction: int
        the direction of the lock operation: 0 or 1 (see above convention)
    """
    direction = 0
    if to_node == lock.start_node:
        direction = 1

    return direction



# packages for time, space and id
import datetime

# packages for data handling
import numpy as np
import pandas as pd

# plotting libraries
import matplotlib.pyplot as plt
from plotly.offline import init_notebook_mode, iplot
import plotly.graph_objs as go

# spatial libraries
import networkx as nx
from simplekml import Kml, Style


def vessel_planning(vessels, activities, colors, web=False):
    """create a plot of the planning of vessels

    Parameters
    ----------
    vessels : list of objects
        A list containing vessel objects to plot. vessel need to have a logbook (mixin class Log)
    activities : list of strings
        The activities to plot. Must be denoted in logbook as '<activity> start' and '<activity> stop'
    colors : list of colors
        The colors to use for the activities. must be the same length as activities
    """

    def get_segments(series, activity, y_val):
        """extract 'start' and 'stop' of activities from log"""
        x = []
        y = []
        for i, v in series.iteritems():
            if v == activity + " start":
                start = i
            if v == activity + " stop":
                x.extend((start, start, i, i, i))
                y.extend((y_val, y_val, y_val, y_val, None))
        return x, y

    # organise logdata into 'dataframes'
    dataframes = []
    for vessel in vessels:
        rename = {"Value": "log_value", "Message": "log_string"}
        df = pd.DataFrame(vessel.logbook)
        df = df.set_index("Timestamp").rename(columns=rename)

        dataframes.append(df)
    df = dataframes[0]

    # prepare traces for each of the activities
    traces = []
    for i, activity in enumerate(activities):
        x_combined = []
        y_combined = []
        for k, df in enumerate(dataframes):
            y_val = vessels[k].name
            x, y = get_segments(df["log_string"], activity=activity, y_val=y_val)
            x_combined.extend(x)
            y_combined.extend(y)
        traces.append(
            go.Scatter(
                name=activity,
                x=x_combined,
                y=y_combined,
                mode="lines",
                hoverinfo="y+name",
                line=dict(color=colors[i], width=10),
                connectgaps=False,
            )
        )

    # prepare layout of figure
    layout = go.Layout(
        title="Vessel planning",
        hovermode="closest",
        legend=dict(x=0, y=-0.2, orientation="h"),
        xaxis=dict(
            title="Time",
            titlefont=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            range=[0, df.index[-1]],
        ),
        yaxis=dict(
            title="Vessels",
            titlefont=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
        ),
    )

    # plot figure
    init_notebook_mode(connected=True)
    fig = go.Figure(data=traces, layout=layout)
    return iplot(fig, filename="news-source")


def vessel_kml(
    env,
    vessels,
    fname="vessel_movements.kml",
    icon="http://maps.google.com/mapfiles/kml/shapes/sailing.png",
    size=1,
    scale=1,
    stepsize=120,
):
    """Create a kml visualisation of vessels and save in defined file.

    Parameters
    ----------
    env : simpy.Environment
        The environment object containing the simulation time. Env variable needs to contain epoch to enable conversion
        of simulation time to real time.
    vessels: list of vessels
        Vessels need logs that contain geometries in lat, lon as a function of time. (mixin class Log)
    fname: str
        The name of the kml file to be created. Default is 'vessel_movements.kml'
    icon: str
        The icon to be used for the vessels. Default is 'http://maps.google.com/mapfiles/kml/shapes/sailing.png'
    size: float
        The size of the label. Default is 1
    scale: float
        The scale of the icon. Default is 1
    stepsize: int
        The stepsize for the interpolation of the geometry. Default is 120 seconds (2 minutes)
    """

    # create a kml file containing the visualisation
    kml = Kml()
    fol = kml.newfolder(name="Vessels")

    shared_style = Style()
    shared_style.labelstyle.color = "ffffffff"  # White
    shared_style.labelstyle.scale = size
    shared_style.iconstyle.color = "ffff0000"  # Blue
    shared_style.iconstyle.scale = scale
    shared_style.iconstyle.icon.href = icon

    # each timestep will be represented as a single point
    # todo: create a tmpvessel to log info. Do not attach it to the original vessel log!
    for vessel in vessels:
        tmp_vessel = {
            "Geometry - x": [],
            "Geometry - y": [],
            "timestamps_t": [],
            "timestamps_x": [],
        }
        geom_x = []
        geom_y = []

        vessel_log = pd.DataFrame(vessel.logbook)

        for geom in vessel_log["Geometry"]:
            geom_x.append(geom.x)
            geom_y.append(geom.y)

        tmp_vessel["Geometry - x"] = geom_x
        tmp_vessel["Geometry - y"] = geom_y

        time_stamp_min = min(vessel_log["Timestamp"]).timestamp()
        time_stamp_max = max(vessel_log["Timestamp"]).timestamp()

        steps = int(np.floor((time_stamp_max - time_stamp_min) / stepsize))
        timestamps_t = np.linspace(time_stamp_min, time_stamp_max, steps)

        times = []
        for row in vessel.logbook:
            t = row["Timestamp"]
            times.append(t.timestamp())

        tmp_vessel["timestamps_t"] = timestamps_t
        tmp_vessel["timestamps_x"] = np.interp(timestamps_t, times, tmp_vessel["Geometry - x"])
        tmp_vessel["timestamps_y"] = np.interp(timestamps_t, times, tmp_vessel["Geometry - y"])

        log_index = -1
        for log_index, value in enumerate(tmp_vessel["timestamps_t"][:-1]):
            begin = datetime.datetime.fromtimestamp(tmp_vessel["timestamps_t"][log_index])
            end = datetime.datetime.fromtimestamp(tmp_vessel["timestamps_t"][log_index + 1])

            pnt = fol.newpoint(
                name=vessel.name,
                coords=[
                    (
                        tmp_vessel["timestamps_x"][log_index],
                        tmp_vessel["timestamps_y"][log_index],
                    )
                ],
            )
            pnt.timespan.begin = begin.isoformat()
            pnt.timespan.end = end.isoformat()
            pnt.style = shared_style

        # include last point as well
        begin = datetime.datetime.fromtimestamp(tmp_vessel["timestamps_t"][log_index + 1])
        # end = datetime.datetime.fromtimestamp(vessel.log["timestamps_t"][log_index + 1])

        pnt = fol.newpoint(
            name=vessel.name,
            coords=[
                (
                    tmp_vessel["timestamps_x"][log_index + 1],
                    tmp_vessel["timestamps_y"][log_index + 1],
                )
            ],
        )
        pnt.timespan.begin = begin.isoformat()
        # pnt.timespan.end = end.isoformat()
        pnt.style = shared_style

    kml.save(fname)


def site_kml(
    env,
    sites,
    fname="site_development.kml",
    icon="http://maps.google.com/mapfiles/kml/shapes/square.png",
    size=1,
    scale=3,
):
    """Create a kml visualisation of vessels.

    Parameters
    ----------
    env : simpy.Environment
        The environment object containing the simulation time. Env variable needs to contain epoch to enable conversion of
        the simulation time to real time.
    sites: list of sites
        Sites need logs that contain geometries in lat, lon as a function of time. (mixin class Log)
    fname: str
        The name of the kml file to be created. Default is 'site_development.kml'
    icon: str
        The icon to be used for the vessels. Default is 'http://maps.google.com/mapfiles/kml/shapes/square.png'
    size: float
        The size of the label. Default is 1
    scale: float
        The scale of the icon. Default is 1
    """

    # create a kml file containing the visualisation
    kml = Kml()
    fol = kml.newfolder(name="Sites")

    # each timestep will be represented as a single point
    for site in sites:
        for log_index, value in enumerate(site.log["Timestamp"][:-1]):
            style = Style()
            style.labelstyle.color = "ffffffff"  # White
            style.labelstyle.scale = size
            style.iconstyle.color = "ff00ffff"  # Yellow
            style.iconstyle.scale = scale * (site.log["Value"][log_index] / site.container.capacity)
            style.iconstyle.icon.href = icon

            begin = site.log["Timestamp"][log_index]
            end = site.log["Timestamp"][log_index + 1]

            pnt = fol.newpoint(
                name=site.name,
                coords=[
                    (
                        site.log["Geometry"][log_index].x,
                        site.log["Geometry"][log_index].y,
                    )
                ],
            )
            pnt.timespan.begin = begin.isoformat()
            pnt.timespan.end = end.isoformat()
            pnt.style = style

        # include last point as well
        style = Style()
        style.labelstyle.color = "ffffffff"  # White
        style.labelstyle.scale = 1
        style.iconstyle.color = "ff00ffff"  # Yellow
        style.iconstyle.scale = scale * (site.log["Value"][log_index + 1] / site.container.capacity)
        style.iconstyle.icon.href = icon

        begin = site.log["Timestamp"][log_index + 1]
        # end = site.log["Timestamp"][log_index + 1]

        pnt = fol.newpoint(
            name=site.name,
            coords=[
                (
                    site.log["Geometry"][log_index + 1].x,
                    site.log["Geometry"][log_index + 1].y,
                )
            ],
        )
        pnt.timespan.begin = begin.isoformat()
        # pnt.timespan.end = end.isoformat()
        pnt.style = style

    kml.save(fname)


def graph_kml(
    env,
    fname="graph.kml",
    icon="http://maps.google.com/mapfiles/kml/shapes/donut.png",
    size=0.5,
    scale=0.5,
    width=5,
):
    """Create a kml visualisation of graph and save in defined file.

    Parameters
    ----------
    env : simpy.Environment
        The environment object containing the simulation time. Env variable needs to contain the graph.
    fname: str
        The name of the kml file to be created. Default is 'graph.kml'
    icon: str
        The icon to be used for the vessels. Default is 'http://maps.google.com/mapfiles/kml/shapes/donut.png'
    size: float
        The size of the label. Default is 0.5
    scale: float
        The scale of the icon. Default is 0.5
    width: float
        The width of the line. Default is 5
    """

    # create a kml file containing the visualisation
    kml = Kml()
    fol = kml.newfolder(name="Vessels")

    shared_style = Style()
    shared_style.labelstyle.color = "ffffffff"  # White
    shared_style.labelstyle.scale = size
    shared_style.iconstyle.color = "ffffffff"  # White
    shared_style.iconstyle.scale = scale
    shared_style.iconstyle.icon.href = icon
    shared_style.linestyle.color = "ff0055ff"  # Red
    shared_style.linestyle.width = width

    nodes = list(env.graph.nodes)

    # each timestep will be represented as a single point
    for log_index, value in enumerate(list(env.graph.nodes)[0 : -1 - 1]):
        pnt = fol.newpoint(
            name="",
            coords=[
                (
                    nx.get_node_attributes(env.graph, "geometry")[nodes[log_index]].x,
                    nx.get_node_attributes(env.graph, "geometry")[nodes[log_index]].y,
                )
            ],
        )
        pnt.style = shared_style

    edges = list(env.graph.edges)
    for log_index, value in enumerate(list(env.graph.edges)[0 : -1 - 1]):
        lne = fol.newlinestring(
            name="",
            coords=[
                (
                    nx.get_node_attributes(env.graph, "geometry")[edges[log_index][0]].x,
                    nx.get_node_attributes(env.graph, "geometry")[edges[log_index][0]].y,
                ),
                (
                    nx.get_node_attributes(env.graph, "geometry")[edges[log_index][1]].x,
                    nx.get_node_attributes(env.graph, "geometry")[edges[log_index][1]].y,
                ),
            ],
        )
        lne.style = shared_style

    kml.save(fname)


def energy_use(vessel, testing=False):
    """Create a plot of the energy use of a vessel

    Parameters
    ----------
    vessel : object
        A vessel object to plot. vessel need to have a logbook (mixin class Log)
    testing : bool
        If True, the plot will not be shown. This is useful for testing purposes.
    """
    energy_use_loading = 0  # concumption between loading start and loading stop
    energy_use_sailing_full = 0  # concumption between sailing full start and sailing full stop
    energy_use_unloading = 0  # concumption between unloading  start and unloading  stop
    energy_use_sailing_empty = 0  # concumption between sailing empty start and sailing empty stop
    energy_use_waiting = 0  # concumption between waiting start and waiting stop

    for i in range(len(vessel.log["Message"])):
        if vessel.log["Message"][i] == "Energy use loading":
            energy_use_loading += vessel.log["Value"][i]

        elif vessel.log["Message"][i] == "Energy use sailing full":
            energy_use_sailing_full += vessel.log["Value"][i]

        elif vessel.log["Message"][i] == "Energy use unloading":
            energy_use_unloading += vessel.log["Value"][i]

        elif vessel.log["Message"][i] == "Energy use sailing empty":
            energy_use_sailing_empty += vessel.log["Value"][i]

        elif vessel.log["Message"][i] == "Energy use waiting":
            energy_use_waiting += vessel.log["Value"][i]

    # For the total plot
    fig, ax1 = plt.subplots(figsize=[15, 10])

    # For the barchart
    height = [
        energy_use_loading,
        energy_use_unloading,
        energy_use_sailing_full,
        energy_use_sailing_empty,
        energy_use_waiting,
    ]
    labels = ["Loading", "Unloading", "Sailing full", "Sailing empty", "Waiting"]
    colors = [
        (55 / 255, 126 / 255, 184 / 255),
        (98 / 255, 192 / 255, 122 / 255),
        (255 / 255, 150 / 255, 0 / 255),
        (98 / 255, 141 / 255, 122 / 255),
        (124 / 255, 10 / 255, 2 / 255),
    ]

    positions = np.arange(len(labels))
    ax1.bar(positions, height, color=colors)

    # For the cumulative percentages
    total_use = sum(
        [
            energy_use_loading,
            energy_use_unloading,
            energy_use_sailing_full,
            energy_use_sailing_empty,
            energy_use_waiting,
        ]
    )

    energy_use_unloading += energy_use_loading
    energy_use_sailing_full += energy_use_unloading
    energy_use_sailing_empty += energy_use_sailing_full
    energy_use_waiting += energy_use_sailing_empty
    y = [
        energy_use_loading,
        energy_use_unloading,
        energy_use_sailing_full,
        energy_use_sailing_empty,
        energy_use_waiting,
    ]
    n = [
        energy_use_loading / total_use,
        energy_use_unloading / total_use,
        energy_use_sailing_full / total_use,
        energy_use_sailing_empty / total_use,
        energy_use_waiting / total_use,
    ]

    ax1.plot(positions, y, "ko", markersize=10)
    ax1.plot(positions, y, "k")

    for i, txt in enumerate(n):
        x_txt = positions[i] + 0.1
        y_txt = y[i] * 0.95
        ax1.annotate("{:02.1f}%".format(txt * 100), (x_txt, y_txt), size=12)

    # Further markup
    plt.ylabel("Energy useage in KWH", size=12)
    ax1.set_xticks(positions)
    ax1.set_xticklabels(labels, size=12)
    plt.title("Energy use - {}".format(vessel.name), size=15)

    if testing is False:
        plt.show()



"""
Core utilities related to plotting.
"""

# %% IMPORT DEPENDENCIES
# generic
import pandas as pd
import plotly.express as px

from plotly.offline import init_notebook_mode, iplot


# %% GENERATE GANTT CHART FROM EVENT TABLE
def generate_vessel_gantt_chart(df_eventtable: pd.DataFrame, static: bool = False):
    """
    Method to generate a Gantt chart from a vessel activity log DataFrame.

    This method visualizes the activity timeline of vessels by combining vessel
    names and activity types into a single label, and plotting them using Plotly
    Express's timeline chart.

    Parameters
    ----------
    df_eventtable : pandas.DataFrame
        DataFrame containing columns 'object name', 'activity name', 'start time',
        and 'stop time' representing vessel activity logs.
    static : bool, optional
        If True, returns a static Plotly figure object.
        If False, displays the figure

    Returns
    -------
    fig : plotly.graph_objects.Figure
        A Plotly figure object representing the Gantt chart of vessel activities.
    """

    # Add vessel name to activity label
    df_eventtable["activity label"] = (
        df_eventtable["object name"] + " - " + df_eventtable["activity name"]
    )

    # Create the Gantt chart
    fig = px.timeline(
        df_eventtable,
        x_start="start time",
        x_end="stop time",
        y="activity label",
        color="object name",
        title="Gantt chart of logged events",
    )

    # Reverse the Y-axis to match Gantt chart style
    fig.update_yaxes(autorange="reversed")

    # Customize layout
    fig.update_layout(
        xaxis_title="Time",
        yaxis_title="Activity (with Vessel)",
        legend_title="Vessel",
        margin=dict(l=20, r=20, t=40, b=20),
    )

    if static is False:
        # Initialize notebook mode for Plotly
        init_notebook_mode(connected=True)
        # Display the figure in a Jupyter notebook
        iplot(fig)
    else:
        return fig



"""
Logging utilities for energy-related calculations.
"""

# %% IMPORT DEPENDENCIES
# generic


# internal
import opentnsim.graph.mixins as graph_module


# %% ADD ENERGY ATTRIBUTES INTO EVENT TABLE
def add_energy_attributes_to_eventtable(df, objs):
    """
    Add energy-related attributes to a 'minimum event table'.

    When a vessel object is created with the ConsumesEnergy and VesselProperties
    mixins, and the path along the graph contains Info - Generaldepth
    information, the energy module can calculate:
     - resistance components for passing an edge at a given speed,
     - power needed to overcome that resistance,
     - energy consumed to deliver that power.

    This method takes a 'minimum event table' and a list of objs as input, and
    returns an event table with resistance, power, and energy attributes.

    References
    ----------
    Van der Werff, S.E., F. Baart and M. van Koningsveld (2025). “Merging
    Multiple System Perspectives: The Key to Effective Inland Shipping
    Emission-Reduction Policy Design.” Journal of Marine Science and Engineering
    13(4), 716. https://doi.org/10.3390/jmse13040716

    Van der Werff, S.E., S. Eppenga, A. van der Hout, F. Baart and M. van
    Koningsveld (2025). “Multi-perspective nautical safety risk assessment of
    allisions with offshore wind parks.” Applied Ocean Research 158(2025),104564.
    https://doi.org/10.1016/j.apor.2025.104564

    For waterborne traffic over a network, a unique event is defined by:
     - a unique vessel,
     - a specific section of the waterway,
     - a specific time.

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame with elements of a 'minimum event table'.
    objs: list
        List of OpenTNSim simulation objects with log information.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame with energy-related attributes added.
    """

    for index, row in df.iterrows():

        # the generator option (next) is used to make sure we get the full copy of
        # the object found in the list
        obj = next((x for x in objs if x.id == row["object id"]), None)

        # get the depth from the edge sailed in the event (and check if squat effects
        # need to be considered
        h_0 = graph_module.calculate_depth(row["start location"], row["stop location"], obj.env.graph)
        h_0 = obj.calculate_h_squat(
            v=obj.v, h_0=h_0
        )  # TODO: actually takes width as arg

        obj.calculate_total_resistance(v=obj.v, h_0=h_0)
        obj.calculate_total_power_required(v=obj.v, h_0=h_0)
        obj.calculate_emission_factors_total(v=obj.v, h_0=h_0)
        obj.calculate_SFC_final(v=obj.v, h_0=h_0)

        df.at[index, "waterdepth (m)"] = h_0
        df.at[index, "waterway width (m)"] = None
        df.at[index, "current (m/s)"] = 0  # TODO: get current from graph
        df.at[index, "engine age (year)"] = obj.C_year

        df.at[index, "P_tot (kW)"] = obj.P_tot
        df.at[index, "P_given (kW)"] = obj.P_given
        df.at[index, "P_installed (kW)"] = obj.P_installed

        energy_delta = obj.P_tot * row["duration (s)"] / 3600  # kJ/3600 = kWh
        df.at[index, "total_energy (kWh)"] = energy_delta

    return df


# %% ADD FUEL ATTRIBUTES INTO EVENT TABLE
def add_fuel_attributes_to_event_table(df, objs):
    """
    Method to add fuel-related attributes to a 'minimum event table' plus
    energy attributes.

    When we have an event table that includes energy attributes, we can add
    fuel-related attributes. Depending on the energy carrier and the energy
    converter, we can estimate:
     - fuel consumption
     - emissions

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame that contains the elements of a 'minimum event table' plus
        energy-related attributes.
    objs: list
        List of OpenTNSim simulation objects that have log information.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame that contains a 'minimum event table' plus energy-related and
        fuel-related attributes.
    """
    # check if total_energy (kWh) is in the DataFrame
    if "total_energy (kWh)" not in df.columns:
        raise ValueError(
            "DataFrame must contain 'total_energy (kWh)' column to add fuel "
            "attributes. Make sure to call add_energy_attributes_to_eventtable first."
        )

    for index, row in df.iterrows():

        # the generator option (next) is used to make sure we get the full copy of the object found in the list
        obj = next((x for x in objs if x.id == row["object id"]), None)

        diesel_C_year = row["total_energy (kWh)"] * obj.final_SFC_diesel_C_year_ICE_mass

        df.at[index, "diesel_consumption (g)"] = diesel_C_year
        df.at[index, "diesel_consumption_m (g/m)"] = diesel_C_year / row["distance (m)"]
        df.at[index, "diesel_consumption_s (g/s)"] = diesel_C_year / row["duration (s)"]

        CO2_emission_total = (
            row["total_energy (kWh)"] * obj.total_factor_CO2
        )  # in g (total_factor is the Emission Factor)                                                                                              #stationary phase # in g
        PM10_emission_total = row["total_energy (kWh)"] * obj.total_factor_PM10  # in g
        NOX_emission_total = row["total_energy (kWh)"] * obj.total_factor_NOX  # in g

        df.at[index, "CO2_emission_total (g)"] = CO2_emission_total
        df.at[index, "PM10_emission_total (g)"] = PM10_emission_total
        df.at[index, "NOX_emission_total (g)"] = NOX_emission_total

        # TODO: see if these entries are useful. They can also be easily calculated
        # from other entries
        df.at[index, "CO2_emission_per_m (g/m)"] = (
            CO2_emission_total / row["distance (m)"]
        )
        df.at[index, "PM10_emission_per_m (g/m)"] = (
            PM10_emission_total / row["distance (m)"]
        )
        df.at[index, "NOX_emission_per_m (g/m)"] = (
            NOX_emission_total / row["distance (m)"]
        )

        # TODO: see if these entries are useful. They can also be easily calculated
        # from other entries
        df.at[index, "CO2_emission_per_s (g/s)"] = (
            CO2_emission_total / row["duration (s)"]
        )
        df.at[index, "PM10_emission_per_s (g/s)"] = (
            PM10_emission_total / row["duration (s)"]
        )
        df.at[index, "NOX_emission_per_s (g/s)"] = (
            NOX_emission_total / row["duration (s)"]
        )

    return df



"""Vessel generator."""

# packkage(s) for documentation, debugging, saving and loading
import datetime
import logging

# package(s) for data handling
import random
import pandas as pd


# package(s) related to time, space and id
import time
import uuid
import numbers
import datetime

# you need these dependencies (you can get these from anaconda)
# package(s) related to the simulation
import simpy
import networkx as nx


# import core from self
import opentnsim.core as core
import opentnsim.vessel_traffic_service.vessel_traffic_service as vessel_traffic_service

logger = logging.getLogger(__name__)


class VesselGenerator:
    """
    A class to generate vessels from a database

    Parameters
    ----------
    vessel_type: class of mixins.
        The type of vessel to be generated.
    vessel_database: ??
        The database from which the vessel is generated.
        Make sure all needed attributes for vessel_type are available in the database.
    loaded: optional
        whether or not the vessel is loaded.
        If True, the vessel is loaded.
        If "Random", the vessel is randomly loaded or not (50% chance fully loaded, 50% chance empty).
        If not specified, the vessel is empty.
    random_seed: int, optional
        The random seed for generating vessels. The default is 3.
    """

    def __init__(self, vessel_type, vessel_database, loaded=None, random_seed=3):
        """Initialization"""

        self.vessel_type = vessel_type
        self.vessel_database = vessel_database
        self.loaded = loaded
        random.seed(random_seed)

    def generate(self, environment, vessel_name: str, fleet_distribution=None, scenario=None):
        """Get a random vessel from self.database

        Parameters
        ----------
        environment: simpy environment
            The environment in which the vessel is generated.
        vessel_name: str
            The name that is assigned to the generated vessel.
        scenario: str, optional
            The scenario of the generated vessel. If given, the vessel with this scenario is selected from the database.
        """

        if fleet_distribution == None:
            vessel_info = self.vessel_database.sample(n=1, random_state=int(1000 * random.random()))
        else:
            vessel_info = self.vessel_database.sample(n=1, weights=fleet_distribution, random_state=int(1000 * random.random()))
        vessel_data = {}

        vessel_data["env"] = environment
        vessel_data["name"] = vessel_name

        if scenario:
            vessel_info = vessel_info[vessel_info["scenario"] == scenario]

        for key in vessel_info:
            if key == "vessel_id":
                vessel_data["id"] = vessel_info[key].values[0]
            else:
                vessel_data[key] = vessel_info[key].values[0]

        if self.loaded is True:
            vessel_data["level"] = vessel_data["capacity"]
        elif self.loaded == "Random":
            if random.random() < 0.5:
                vessel_data["level"] = vessel_data["capacity"]
            else:
                vessel_data["level"] = 0
        vessel_data["route"] = None
        vessel_data["geometry"] = None
        self.vessel_type(**vessel_data)
        return self.vessel_type(**vessel_data)

    def arrival_process(
        self,
        environment,
        origin,
        destination,
        arrival_distribution,
        scenario,
        arrival_process,
        fleet_distribution,
    ):
        """
        Make arrival process in the simulation environment. Vessels with a route
        and between origin and destination are generated according to the arrival distribution.
        The route is calculated using the dijkstra algorithm.

        Parameters
        ----------
        environment:  simpy environment
            The environment in which the vessel is generated.
        origin:       str
            The origin of the vessel.
        destination:  str
            The destination of the vessel.
        arrival_distribution: int or list
            The amount of vessels that enter the simulation per hour.
            If int, it is the average number of vessels per hour over the entire day.
            If list, it is the average number of vessels per hour for each hour of the day.
            List must have length of 24 (one entry for each hour).
        scenario:     str
            The scenario that is assigned to the generated vessel.
        arrival_process:  str
            process of arrivals. choose between "Markovian" or "Uniform".
        """

        # Create an array with inter-arrival times -- average number of seconds between arrivals
        if type(arrival_distribution) == int or type(arrival_distribution) == float:
            self.inter_arrival_times = [3600 / arrival_distribution] * 24

        elif type(arrival_distribution) == list and len(arrival_distribution) == 24:
            self.inter_arrival_times = [3600 / n for n in arrival_distribution]

        elif type(arrival_distribution) == list:
            raise ValueError("List should contain an average number of vessels per hour for an entire day: 24 entries.")

        else:
            raise ValueError("Specify an arrival distribution: type Integer or type List.")

        while True:

            # Check simulation time
            inter_arrival = self.inter_arrival_times[datetime.datetime.fromtimestamp(environment.now).hour]

            # In the case of a Markovian arrival process
            if arrival_process == "Markovian":

                # Make a timestep based on the poisson process
                yield environment.timeout(random.expovariate(1 / inter_arrival))

            elif arrival_process == "Uniform":

                # Make a timestep based on uniform arrivals
                yield environment.timeout(inter_arrival)

            else:
                raise ValueError(
                    "No other arrival processes are yet defined. You can add them to "
                    "transport_network_analysis/vessel_generator.py."
                )

            # Create a vessel
            vessel = self.generate(environment, "Vessel", fleet_distribution, scenario)
            vessel.output = {}
            # core.Output.vessel_dependent_output(vessel)
            vessel.env = environment
            vessel.route = nx.dijkstra_path(environment.graph, origin, destination)
            vessel.geometry = nx.get_node_attributes(environment.graph, "geometry")[vessel.route[0]]
            environment.vessels.append(vessel)
            # Move on path
            process = environment.process(vessel.move())
            vessel.process = process


class Simulation(core.Identifiable):
    """
    A class to generate vessels from a database

    Parameters
    ----------
    simulation_start: datetime
        The start time of the simulation.
    graph: networkx graph
        The graph that is used for the simulation.
    scenario:
        scenario with vessels - should be coupled to the database
    """

    def __init__(self,
                 graph,
                 simulation_start=datetime.datetime.now(),
                 simulation_duration=None,
                 simulation_stop=None,
                 hydrodynamic_start_time=datetime.datetime.now(),
                 hydrodynamic_data_path=None,
                 vessel_speed_data_path=None, scenario=None):
        """ 
        Initialization 
        
        scenario: scenario with vessels - should be coupled to the database
        """
        self.env = simpy.Environment(initial_time=time.mktime(simulation_start.timetuple()))
        self.env.graph = graph
        self.env.simulation_start = simulation_start
        self.env.simulation_stop = simulation_stop
        self.simulation_duration = simulation_duration
        if self.env.simulation_stop:
            self.simulation_duration = self.env.simulation_stop-self.env.simulation_start
        else:
            self.env.simulation_stop = self.env.simulation_start+self.simulation_duration
        self.env.routes = pd.DataFrame.from_dict(
            {
                "Origin": [],
                "Destination": [],
                "Width": [],
                "Height": [],
                "Depth": [],
                "Route": [],
            }
        )
        self.scenario = scenario

        self.env.vessels = []
        self.output = {}

        self.env.vessel_traffic_service = vessel_traffic_service.VesselTrafficService(FG=graph,
                                                                                      hydrodynamic_start_time = hydrodynamic_start_time,
                                                                                      hydrodynamic_information_path = hydrodynamic_data_path,
                                                                                      vessel_speed_data_path=vessel_speed_data_path)

        self.environment.vessel_traffic_service = vessel_traffic_service.VesselTrafficService(hydrodynamic_data, vessel_speed_data)

    def add_vessels(
        self,
        origin=None,
        destination=None,
        vessel=None,
        vessel_generator: VesselGenerator = None,
        fleet_distribution=None,
        arrival_distribution=1,
        arrival_process="Markovian",
    ):
        """
        Make arrival process in the environment. Add one vessel  on one input vessel, or with the vessel generator

        Parameters
        ----------
        vessel_generator:      VesselGenerator
            The vessel generator that is used to generate vessels. Optional. If not specified,
            the vessel should be specified.
            If specified, the vessel should be None, but origin, destination, arrival_distribution
            and arrival_process should be specified.
        origin: str
            The origin of the vessel. Must be specified if vessel_generator is specified.
        destination: str
            The destination of the vessel. Must be specified if vessel_generator is specified.
        arrival_distribution: int or list
            The amount of vessels that enter the simulation per hour. Must be specified if
            vessel_generator is specified.
            If int, it is the average number of vessels per hour over the entire day.
            If list, it is the average number of vessels per hour for each hour of the day.
            List must have length of 24 (one entry for each hour).
        arrival_process: str
            The process of arrivals. Must be specified if vessel_generator is specified.
            Choose between "Markovian" or "Uniform".
        vessel: Vessel
            A vessel object with a route between origin and destination. Optional.
            If specified, the vessel_generator should be None, and origin, destination,
            arrival_distribution and arrival_process are ignored.
        """

        if vessel_generator == None:
            self.env.vessels.append(vessel)
            process = self.env.process(vessel.move())
            vessel.process = process
            if 'metadata' in dir(vessel) and 'arrival_time' not in vessel.metadata.keys():
                vessel.metadata["arrival_time"] = self.env.simulation_start

        else:
            self.env.process(
                vessel_generator.arrival_process(
                    self.env,
                    origin,
                    destination,
                    arrival_distribution,
                    self.scenario,
                    arrival_process,
                    fleet_distribution,
                )
            )

    def run(self, duration=24 * 60 * 60):
        """
        Run the simulation

        duration:   float
            specify the duration of the simulation in seconds
        """
        self.env.run(until=self.env.now + self.simulation_duration.total_seconds())



# package(s) related to time, space and id
import bisect
import scipy as sc
from shapely.geometry import Point, LineString, MultiLineString
from shapely.ops import linemerge
import shapely
import pandas as pd
from matplotlib import pyplot as plt, dates
import math
import networkx as nx
import datetime
import pytz
import pyproj

# mathematical packages
import bisect
import math
import scipy as sc

# packages for data handling
import numpy as np
import pandas as pd
import xarray as xr
import pickle
# import netCDF4
from shapely import reverse
from shapely.ops import transform
from opentnsim import core
from opentnsim import model
from opentnsim.graph.mixins import get_length_of_edge, get_geometry_of_edge, determine_length_of_edge_geometry
from opentnsim.graph import mixins
from opentnsim.vessel_traffic_service.hydrodanamic_data_manager import HydrodynamicDataManager

# spatial libraries
import networkx as nx
import shapely
import shapely.ops
from shapely.geometry import MultiPolygon, Point, Polygon


class VesselTrafficService(mixins.HasMultiDiGraph):
    """Class: a collection of functions that processes requests of vessels regarding the nautical processes on ow to enter the port safely"""

    def __init__(
        self,
        graph,
        crs_m = "EPSG:4087",
        hydrodynamic_information_path=None,
        vessel_speed_information_path=None,
        hydrodynamic_information=None,
        vessel_speed_information=None,
        hydrodynamic_start_time=pd.Timedelta(seconds=0),
        *args,
        **kwargs,
    ):
        self.crs_m = crs_m
        self.hydrodynamic_information = hydrodynamic_information
        self.hydrodynamic_information_path = hydrodynamic_information_path
        self.graph = graph

        if isinstance(hydrodynamic_information, xr.Dataset):
            self.hydrodynamic_information_path = False
        if isinstance(vessel_speed_information, xr.Dataset):
            self.vessel_speeds = vessel_speed_information

        global vertical_tidal_restrictions_condition_df
        self.vertical_tidal_restrictions_condition_df = pd.DataFrame()

        global horizontal_tidal_restrictions_condition_df
        self.horizontal_tidal_restrictions_condition_df = pd.DataFrame()

        global restricted_vessel_speeds
        self.restricted_vessel_speeds = pd.DataFrame()

        global edges_info
        self.edges_info = self.get_edges_info()

        for edge in self.graph.edges:
            length = determine_length_of_edge_geometry(graph, edge, crs_meter = self.crs_m)
            self.graph.edges[edge]["length_m"] = length

        for node in graph.nodes:
            node_info = graph.nodes[node]
            if 'Horizontal tidal restriction' in node_info.keys():
                specification_df = graph.nodes[node]['Horizontal tidal restriction']['Specifications']
                specification_df['Node'] = node
                self.horizontal_tidal_restrictions_condition_df = pd.concat(
                    [self.horizontal_tidal_restrictions_condition_df, specification_df]
                )
            if 'Vertical tidal restriction' in node_info.keys():
                specification_df = graph.nodes[node]['Vertical tidal restriction']['Specifications']
                specification_df['Node'] = node
                self.vertical_tidal_restrictions_condition_df = pd.concat(
                    [self.vertical_tidal_restrictions_condition_df, specification_df]
                )

        self.horizontal_tidal_restrictions_condition_df = self.horizontal_tidal_restrictions_condition_df.reset_index(drop=True)
        self.vertical_tidal_restrictions_condition_df = self.vertical_tidal_restrictions_condition_df.reset_index(drop=True)

        if self.hydrodynamic_information_path is not None:
            hydro_manager = HydrodynamicDataManager()
            if isinstance(hydrodynamic_information_path,str):
                hydro_manager.hydrodynamic_data = netCDF4.Dataset(self.hydrodynamic_information_path)
            else:
                hydro_manager.hydrodynamic_data = hydrodynamic_information

            self.hydrodynamic_start_time = hydrodynamic_start_time
            if isinstance(hydrodynamic_information_path, str):
                self.hydrodynamic_times = hydro_manager.hydrodynamic_times = (
                    hydro_manager.hydrodynamic_data["TIME"][:].data.astype("timedelta64[m]") + hydrodynamic_start_time
                )
            else:
                self.hydrodynamic_times = hydro_manager.hydrodynamic_data["TIME"][:]

        if isinstance(vessel_speed_information_path, str):
            with open(vessel_speed_information_path, "rb") as file:
                self.restricted_vessel_speeds = pickle.load(file)

        super().__init__(*args, **kwargs)

    def get_edges_info(self):
        graph = self.multidigraph
        edges_info = pd.DataFrame(columns=["Edge", "Distance", "MBL"])
        for edge in graph.edges:
            edge_info = graph.edges[edge]
            index = len(edges_info)
            edges_info.loc[index, "Edge"] = edge
            edges_info.loc[index, "Distance"] = get_length_of_edge(graph, edge)
            if "MBL" in edge_info.keys():
                edges_info.loc[index, "MBL"] = edge_info["MBL"]
            else:
                edges_info.loc[index, "MBL"] = 999.0
        return edges_info.set_index("Edge")

    def read_tidal_periods(self,hydrodynamic_data,tidal_period_type,station_index):
        if 'tidal_period_type' not in hydrodynamic_data.variables:
            return None
        data = hydrodynamic_data[tidal_period_type][station_index, :, :]
        for tide_index, (time_start, tide) in enumerate(data):
            if time_start == 'nan':
                new_time = 'NaT'
            else:
                new_time = time_start
            data[tide_index] = (np.datetime64(new_time), tide)
        return data

    def provide_waiting_time_for_inbound_tidal_window(self, vessel, route, time_start=None, time_stop=None, delay=0, plot=False):
        """Function: calculates the time that a vessel has to wait depending on the available tidal windows

        Input:
            - vessel: an identity which is Identifiable, Movable, and Routeable, and has VesselProperties
            - route: a list of strings that resemble the route of the vessel (can be different than the vessel.route)
            - delay: a delay that can be included to calculate a future situation
            - plot: bool that specifies if a plot is requested or not

        """

        # Create sub-routes based on anchorage areas on the route
        if not time_start:
            time_start = pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now)).to_datetime64()
        if not time_stop:
            time_stop = pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now + vessel.metadata['max_waiting_time'])).to_datetime64()
        _,tidal_windows = self.provide_tidal_windows(vessel,route,time_start,time_stop,delay,plot=plot)

        waiting_time = pd.Timedelta(vessel.metadata["max_waiting_time"], "s")
        for window in tidal_windows:
            if time_start > window[1]:
                continue
            if time_start >= window[0]:
                waiting_time = pd.Timedelta(0, "s")
            else:
                waiting_time = window[0] - time_start
            break

        waiting_time = waiting_time.total_seconds()
        return waiting_time

    def provide_waiting_time_for_outbound_tidal_window(self,vessel,route,delay=0,plot=False):
        vessel.bound = 'outbound'
        vessel._T += vessel.metadata['loading'][0]
        waiting_time = self.provide_waiting_time_for_inbound_tidal_window(vessel,route=route,delay=delay, plot=plot)
        vessel._T -= vessel.metadata['loading'][0]
        vessel.bound = 'inbound'
        return waiting_time

    def provide_speed_over_edge(self,vessel,edge):
        v = vessel.v
        restricted_vessel_speeds_edge = self.restricted_vessel_speeds[self.restricted_vessel_speeds.index.isin([edge])]
        if not restricted_vessel_speeds_edge.empty:
            v = restricted_vessel_speeds_edge.Speed.iloc[0]
        if math.isnan(v):
            v = vessel.v
        if 'restricted_speed' in dir(vessel):
            v = vessel.restricted_speed
        return v

    def provide_speed_over_route(self,vessel,route,edges=[]):
        """
        Provides the speed along a vessel's route

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        route : list of str
            str resemble node names that have to be in the graph
        edges : list of tuple
            tuples resemble edges with: a start_node [u] as str, end_node (v) as str, and identifier (k) as int

        Returns
        vessel_speed_over_route : pd.DataFrame
            vessel speed per edge of the route
        -------
        """

        # get edges of route
        if not edges:
            for idx, (u, v) in enumerate(zip(route[:-1], route[1:])):
                k = sorted(self.multidigraph[u][v], key=lambda x: get_length_of_edge(self.multidigraph,(u, v, x)))[0]
                edges.append((u,v,k))

        # construct dataframe of speed information per edge
        vessel_speed_over_route = pd.DataFrame(columns=['Speed'],index=edges)

        # predefined speed restrictions should be added to the dataframe
        if not self.restricted_vessel_speeds.empty:
            vessel_speed_over_route = self.restricted_vessel_speeds[self.restricted_vessel_speeds.index.isin(edges)]
        vessel_speed_over_route = vessel_speed_over_route.reindex(edges)

        # missing speeds should be replaced with the vessel target speed
        vessel_speed_over_route[vessel_speed_over_route.Speed.isna() | (vessel_speed_over_route.Speed == 0)] = vessel.v

        # speeds overruled by lock masters should be obeyed to
        if 'overruled_speed' in dir(vessel):
            for edge,overruled_speed_limit in vessel.overruled_speed.iterrows():
                vessel_speed_over_route.loc[edge,'Speed'] = overruled_speed_limit

        return vessel_speed_over_route

    def provide_heading(self, vessel, edge):
        distance = []
        origin_location = vessel.multidigraph.nodes[edge[0]]["geometry"]
        k = sorted(vessel.multidigraph[edge[0]][edge[1]], key=lambda x: get_length_of_edge(self.multidigraph,(edge[0], edge[1], x)))[0]
        edge_geometry = vessel.multidigraph.edges[edge[0], edge[1], k]["Info"]["geometry"]
        for coord in edge_geometry.coords:
            distance.append(origin_location.distance(Point(coord)))
        if np.argmin(distance):
            edge_geometry = shapely.ops.transform(self.reverse_geometry, edge_geometry)
        heading = np.degrees(
            math.atan2(
                edge_geometry.coords[0][0] - edge_geometry.coords[-1][0], edge_geometry.coords[0][1] - edge_geometry.coords[-1][1]
            )
        )
        return heading

    def reverse_geometry(self, geometry):
        reversed_geometry = reverse(geometry)
        return reversed_geometry

    def provide_trajectory(self,node_1,node_2):
        geometry = None
        route = nx.dijkstra_path(self.multidigraph, node_1, node_2)
        for node_I, node_II in zip(route[:-1], route[1:]):
            k = sorted(self.multidigraph[node_I][node_II], key=lambda x: get_length_of_edge(self.multidigraph,(node_I, node_II, x)))[0]
            edge = (node_I, node_II, k)
            edge_geometry = self.multidigraph.edges[edge]['geometry']
            aligned = self.check_if_geometry_is_aligned_with_edge(edge)
            if not aligned:
                edge_geometry = self.reverse_geometry(edge_geometry)

            if geometry:
                geometry = shapely.ops.linemerge(MultiLineString([geometry, edge_geometry]))
            else:
                geometry = edge_geometry

        return geometry

    def provide_distance_over_network_to_location(self,node_1,node_2,location,tolerance=0.0001):
        geod = pyproj.Geod(ellps="WGS84")
        geometry = self.provide_trajectory(node_1,node_2)
        geometries = shapely.ops.split(shapely.ops.snap(geometry, location, tolerance=tolerance), location).geoms
        distance_sailed = 0
        distance_to_go = 0
        if len(geometries) < 2:
            if self.multidigraph.nodes[node_1]['geometry'] == location:
                distance_to_go = geod.geometry_length(geometries[0])
            elif self.multidigraph.nodes[node_2]['geometry'] == location:
                distance_sailed = geod.geometry_length(geometries[0])
            elif self.multidigraph.nodes[node_1]['geometry'].distance(location) > self.multidigraph.nodes[node_2]['geometry'].distance(location):
                distance_sailed = geod.geometry_length(geometries[0])
            else:
                distance_to_go = geod.geometry_length(geometries[0])
        else:
            distance_sailed = geod.geometry_length(geometries[0])
            distance_to_go = geod.geometry_length(geometries[1])
        return distance_sailed,distance_to_go

    def check_if_geometry_is_aligned_with_edge(self, edge):
        node_start = edge[0]
        node_stop = edge[1]
        edge_geometry = get_geometry_of_edge(self.multidigraph, edge)
        first_point = Point(edge_geometry.coords[0])
        distance_to_edge_nodes = {}
        for node in [node_start, node_stop]:
            node_geometry = self.multidigraph.nodes[node]["geometry"]
            distance_to_edge_nodes[node] = first_point.distance(node_geometry)
        closest_node = min(distance_to_edge_nodes, key=distance_to_edge_nodes.get)
        aligned = closest_node == node_start
        return aligned

    def get_closest_location_on_edge_to_point(self, graph, edge, point):
        edge_geometry = graph.edges[edge]["geometry"]
        point_on_edge = edge_geometry.interpolate(edge_geometry.project(point))
        return point_on_edge

    def transform_geometry(self, geometry, epsg_in="EPSG:4326", epsg_out=None):
        if epsg_out is None:
            epsg_out = self.crs_m
        crs_in = pyproj.CRS(epsg_in)
        crs_out = pyproj.CRS(epsg_out)
        crs_in_to_crs_out = pyproj.transformer.Transformer.from_crs(crs_in, crs_out, always_xy=True).transform
        geometry_transformed = transform(crs_in_to_crs_out, geometry)
        return geometry_transformed

    def provide_location_over_edges(self,node_1,node_2,interpolation_length):
        geometry = self.provide_trajectory(node_1, node_2)
        if geometry is None or geometry.is_empty:
            return None
        geometry_m = self.transform_geometry(geometry, epsg_out = self.crs_m)
        interpolation_point_m = geometry_m.line_interpolate_point(interpolation_length)
        interpolation_point = self.transform_geometry(interpolation_point_m, epsg_in = self.crs_m, epsg_out = "EPSG:4326")
        return interpolation_point

    def provide_distance_from_location_over_edge(self,edge,location,tolerance=0.0001):
        geod = pyproj.Geod(ellps="WGS84")
        if len(edge) == 2:
            k = sorted(self.multidigraph[edge[0]][edge[1]],key=lambda x: get_length_of_edge(self.multidigraph,(edge[0], edge[1], x)))[0]
            edge = (edge[0],edge[1],k)
        geometry = self.multidigraph.edges[(edge[0],edge[1],edge[2])]['geometry']

        distance_sailed = 0
        distance_to_go = 0
        if geometry.coords[0] == location.coords[0]:
            distance_to_go = self.multidigraph.edges[(edge[0],edge[1],edge[2])]['length']
        elif geometry.coords[-1] == location.coords[0]:
            distance_sailed = self.multidigraph.edges[(edge[0],edge[1],edge[2])]['length']
        else:
            lines = shapely.ops.split(shapely.ops.snap(geometry, location, tolerance), location).geoms
            for index, line in enumerate(lines):
                distance = 0
                for point_I, point_II in zip(line.coords[:-1], line.coords[1:]):
                    sub_edge_geometry = LineString([Point(point_I), Point(point_II)])
                    distance += geod.geometry_length(sub_edge_geometry)
                if not index:
                    distance_sailed = distance
                else:
                    distance_to_go = distance
        return distance_sailed, distance_to_go

    def provide_edge_by_distance_from_node(self,env,node_1,node_2,distance):
        route = nx.dijkstra_path(self.multidigraph, node_1, node_2)
        total_length = 0
        for node_I, node_II in zip(route[:-1], route[1:]):
            k = sorted(self.multidigraph[node_I][node_II],key=lambda x: get_length_of_edge(self.multidigraph,(node_I,node_II,x)))[0]
            edge_length = self.multidigraph.edges[node_I,node_II,k]['length']
            total_length += edge_length
            if total_length < distance:
                continue
            break
        return (node_I,node_II,k)

    def provide_sailing_distance(self,edge,k=None):
        """
        Calculates sailing distance of edge

        Parameters
        ----------
        edge : tuple
            tuple resembles an edge with: a start_node [u] as str, end_node (v) as str
        k : int [optional]
            identifier of the edge (used in networkx MultiDiGraph to distinguish between edges between the same pair of nodes)

        Returns
        -------
        sailing_distance : float
            sailing distance along the edge in [m]
        """
        if k is None:
            k = sorted(self.env.multidigraph[edge[0]][edge[1]], key=lambda x: get_length_of_edge(self.multidigraph,(edge[0],edge[1],x)))[0]

        sailing_distance = self.env.multidigraph.edges[edge[0], edge[1], k]['length']

        return sailing_distance

    def provide_sailing_distance_over_route(self, route, edges=None):
        """
        Calculates sailing distance of a route

        Parameters
        ----------
        vessel :
            a type including the following parent-classes: Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        edge : tuple
            tuple resembles an edge with: a start_node [u] as str, end_node (v) as str

        Returns
        -------
        sailing_distance_over_route : float
            sailing distance along the route in [m]
        """

        # get edge info
        edges_info = self.edges_info

        # get edges of route
        if not edges:
            edges = []
            for idx, (u, v) in enumerate(zip(route[:-1], route[1:])):
                k = sorted(self.multidigraph[u][v], key=lambda x: get_length_of_edge(self.multidigraph,(u, v, x)))[0]
                edges.append((u,v,k))

        # calculate sailing distance along route
        sailing_distance_over_route = edges_info[edges_info.index.isin(edges)]

        return sailing_distance_over_route

    def provide_sailing_time(self, vessel, route, edges=None):
        """
        Calculates sailing time of vessel

        Parameters
        ----------
        vessel :
            a type including the following parent-classes: Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        route : list of str
            str resemble node names that have to be in the graph
        edges : list of tuples
            tuples resemble edges with: a start_node [u] as str, end_node (v) as str, and identifier (k) as int

        Returns
        -------
        sailing_time_over_route : pd.DataFrame
            dataframe with edges as (multi)index and the following column-information: Speed, Distance, Time

        """

        # determine list of edges that the vessel passes given its route
        if not edges:
            edges = []
            for idx, (u, v) in enumerate(zip(route[:-1], route[1:])):
                k = sorted(self.multidigraph[u][v], key=lambda x: get_length_of_edge(self.multidigraph,(u, v, x)))[0]
                edges.append((u,v,k))

        # calculate sailing distance over route
        sailing_distance_over_route_df = self.provide_sailing_distance_over_route(route, edges)

        # obtain dataframe with information of sailing speed along route
        sailing_information_df = self.provide_speed_over_route(vessel, route, edges)

        # add distance and duraction information to the sailing information dataframe
        sailing_information_df['Distance'] = sailing_distance_over_route_df['Distance']
        sailing_information_df['Time'] = sailing_information_df['Distance']/sailing_information_df['Speed']

        return sailing_information_df

    def provide_sailing_time_distance_on_edge_to_distance_on_another_edge(self, vessel, route, distance_sailed_on_first_edge=0., distance_sailed_on_last_edge=0., edges=None):
        """
        Calculates the distance from a location along an edge A to another location along an edge B

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        route : list of str
            str resemble node names that have to be in the graph
        distance_sailed_on_first_edge : float
            distance that is already covered on the edge at which the vessel is currently sailing
        distance_sailed_on_last_edge : float
            distance on the last edge that the vessel has to sail to reach its location of interest
        edges : list of tuples
            tuples resemble edges with: a start_node [u] as str, end_node (v) as str, and identifier (k) as int

        Returns
        -------
        sailing_information_df : pd.DataFrame
            dataframe with edges as (multi)index and the following column-information: Speed, Distance, Time

        """

        # obtain dataframe with information of sailing speed, distance and time along route
        sailing_information_df = self.provide_sailing_time(vessel=vessel, route=route, edges=edges)

        # determine indexes of first and last edges
        index_first_edge = pd.Index([sailing_information_df.iloc[0].name])
        index_last_edge = pd.Index([sailing_information_df.iloc[-1].name])

        # determine distance that must still be sailed on the current edge of the vessel
        distance_to_sail_on_first_edge = (sailing_information_df.loc[index_first_edge, 'Distance']-distance_sailed_on_first_edge)

        # adjust information of the sailing distance and sailing time on the first and last edges
        sailing_information_df.loc[index_first_edge, 'Time'] = sailing_information_df.loc[index_first_edge, 'Time']*(distance_to_sail_on_first_edge/sailing_information_df.loc[index_first_edge, 'Distance'])
        sailing_information_df.loc[index_first_edge, 'Distance'] = distance_to_sail_on_first_edge
        sailing_information_df.loc[index_last_edge, 'Time'] = sailing_information_df.loc[index_last_edge, 'Time']*(distance_sailed_on_last_edge/sailing_information_df.loc[index_last_edge, 'Distance'])
        sailing_information_df.loc[index_last_edge, 'Distance'] = distance_sailed_on_last_edge

        return sailing_information_df

    def provide_nearest_anchorage_area(self, vessel, node):
        nodes_of_anchorages = []
        capacity_of_anchorages = []
        users_of_anchorages = []
        sailing_times_to_anchorages = []
        # Loop over the nodes of the network and identify all the anchorage areas:
        for node_anchorage in vessel.multidigraph.nodes:
            if "Anchorage" in vessel.multidigraph.nodes[node_anchorage]:
                # Determine if the anchorage area can be reached
                anchorage_reachable = True
                route_to_anchorage = nx.dijkstra_path(vessel.multidigraph, node, node_anchorage)
                for node_on_route in route_to_anchorage:
                    station_index = list(HydrodynamicDataManager().hydrodynamic_data["STATION"][:]).index(node_on_route)
                    min_water_level = np.min(HydrodynamicDataManager().hydrodynamic_data["Water level"][:, station_index].data)
                    _, _, _, required_water_depth, _, MBL = self.provide_ukc_clearance(vessel,node)
                    if min_water_level - MBL < required_water_depth:
                        anchorage_reachable = False
                        break

                if not anchorage_reachable:
                    continue

                # Extract information over the individual anchorage areas:
                # capacity, users, and the sailing distance to the anchorage area
                # from the designated terminal the vessel is planning to call
                nodes_of_anchorages.append(node_anchorage)
                capacity_of_anchorages.append(vessel.multidigraph.nodes[node_anchorage]["Anchorage"][0].resource.capacity)
                users_of_anchorages.append(len(vessel.multidigraph.nodes[node_anchorage]["Anchorage"][0].resource.users))
                route_from_anchorage = nx.dijkstra_path(vessel.multidigraph, node_anchorage, vessel.route[-1])
                sailing_time_to_anchorage = vessel.env.vessel_traffic_service.provide_sailing_time(vessel, route_from_anchorage)[
                    "Time"
                ].sum()
                sailing_times_to_anchorages.append(sailing_time_to_anchorage)

        # Sort the lists based on the sailing distance to the anchorage area from the designated terminal
        #  the vessel is planning to call
        sorted_nodes_anchorages = [nodes for (distances, nodes) in sorted(zip(sailing_times_to_anchorages, nodes_of_anchorages))]
        sorted_users_of_anchorages = [nodes for (distances, nodes) in sorted(zip(sailing_times_to_anchorages, users_of_anchorages))]
        sorted_capacity_of_anchorages = [
            nodes for (distances, nodes) in sorted(zip(sailing_times_to_anchorages, capacity_of_anchorages))
        ]

        # Take the anchorage area that is closest to the designated terminal the vessel is planning to call if there
        # is sufficient capacity:
        node_anchorage = 0
        for anchorage_index, node_anchorage in enumerate(sorted_nodes_anchorages):
            if sorted_users_of_anchorages[anchorage_index] < sorted_capacity_of_anchorages[anchorage_index]:
                # node anchorage is found
                break
        return node_anchorage

    def provide_governing_current_velocity(self,vessel,node,time_start_index,time_end_index):
        hydromanager = HydrodynamicDataManager()
        station_index = list(hydromanager.hydrodynamic_data["STATION"][:]).index(node)
        times = hydromanager.hydrodynamic_times[time_start_index:time_end_index]
        start_time = times[0]
        end_time = times[-1]
        time_step = times[1]-times[0]
        relative_layer_height = hydromanager.hydrodynamic_data["LAYER"][:].data
        current_velocity = hydromanager.hydrodynamic_data["Primary current velocity"][
            time_start_index:time_end_index, station_index
        ].data

        def depth_averaged_current_velocity(interpolation_depth, times, relative_layer_height, current_velocity, station_index):
            layer_boundaries = []
            average_current_velocity = []
            number_of_layers = len(relative_layer_height)
            water_depth = (
                -1 * hydromanager.hydrodynamic_data["MBL"][:, station_index].data
                + hydromanager.hydrodynamic_data["Water level"][:, station_index].data
            )
            relative_water_depth = np.outer(water_depth, relative_layer_height)
            cumulative_water_depth = np.cumsum(relative_water_depth, axis=1)

            for ti in range(len(times)):
                layer_boundaries.append(
                    np.interp(interpolation_depth, cumulative_water_depth[ti], np.arange(0, number_of_layers, 1))
                )

            layer_boundary = np.floor(layer_boundaries)
            relative_boundary_layer_thickness = layer_boundaries - layer_boundary

            for ti in range(len(times)):
                if int(layer_boundary[ti]) + 2 < len(relative_layer_height):
                    rel_layer_heights = relative_layer_height[0 : int(layer_boundary[ti]) + 2].copy()
                    rel_layer_heights[-1] = rel_layer_heights[-1] * relative_boundary_layer_thickness[ti]
                    average_current_velocity.append(
                        np.average(current_velocity[ti][0 : int(layer_boundary[ti]) + 2], weights=rel_layer_heights)
                    )
                elif int(layer_boundary[ti]) == 0:
                    average_current_velocity = current_velocity[ti]
                else:
                    average_current_velocity.append(np.average(current_velocity[ti], weights=relative_layer_height))

            return average_current_velocity

        if "LAYER" in list(hydromanager.hydrodynamic_data["Current velocity"].dimensions):
            if vessel._T <= 5:
                current_velocity = depth_averaged_current_velocity(5, times, relative_layer_height, current_velocity, station_index)
            elif vessel._T <= 15:
                current_velocity = depth_averaged_current_velocity(
                    vessel._T, times, relative_layer_height, current_velocity, station_index
                )
            else:
                current_velocity = [np.average(current_velocity[t], weights=relative_layer_height) for t in range(len(times))]

        if len(current_velocity) > 2:
            current_governing_current_velocity = current_velocity[2]
        else:
            current_governing_current_velocity = current_velocity[-1]

        return current_velocity, current_governing_current_velocity

    # Functions used to calculate the sail-in-times for a specific vessel

    def provide_tidal_window_restriction(self, vessel, route, node, delay, restriction_type='Vertical'):
        """Function: determines which tidal window restriction applies to the vessel at the specific node

        Input:
            - vessel: an identity which is Identifiable, Movable, and Routable, and has VesselProperties
            - route: a list of strings of node names that resemble the route that the vessel is planning
            to sail (can be different than vessel.route)
            - specifications: the specific data regarding the properties for which the restriction holds
            - node: a string that defines the node of the tidal window restriction

        """

        # Predefined bool
        restriction = []
        boolean = True
        no_tidal_window = True
        node_index = route.index(node)

        # Vessel parameters
        previous_terminal_of_call = np.nan
        terminal_of_call = np.nan
        if len(dict.get(vessel.output, 'visited_terminals')):
            previous_terminal_of_call = dict.get(vessel.output, 'visited_terminals')[-1]
        if len(dict.get(vessel.metadata, 'terminal_of_call')):
            terminal_of_call = dict.get(vessel.metadata, 'terminal_of_call')[0]
        route_before_node = route[:node_index]
        route_after_node = route[node_index:]
        ukc = 0.
        if restriction_type == 'Horizontal':
            restriction_condition_df = self.horizontal_tidal_restrictions_condition_df
            ukc, _, _, _, _, _ = self.provide_ukc_clearance(vessel, node, delay)
        elif restriction_type == 'Vertical':
            restriction_condition_df = self.vertical_tidal_restrictions_condition_df

        node_mask = restriction_condition_df.Node == node
        length_mask = ((restriction_condition_df.min_ge_Length <= vessel.L)&
                       (restriction_condition_df.min_gt_Length < vessel.L)&
                       (restriction_condition_df.max_lt_Length > vessel.L) &
                       (restriction_condition_df.max_le_Length >= vessel.L))
        draught_mask = ((restriction_condition_df.min_ge_Draught <= vessel.T)&
                        (restriction_condition_df.min_gt_Draught < vessel.T)&
                        (restriction_condition_df.max_lt_Draught > vessel.T)&
                        (restriction_condition_df.max_le_Draught >= vessel.T))
        ukc_mask = ((restriction_condition_df.min_ge_UKC <= ukc)&
                    (restriction_condition_df.min_gt_UKC < ukc)&
                    (restriction_condition_df.max_lt_UKC > ukc)&
                    (restriction_condition_df.max_le_UKC >= ukc))
        from_node_mask = (restriction_condition_df.bound_from.isin(route_before_node)|
                          (restriction_condition_df.bound_from.isna()))
        to_node_mask = (restriction_condition_df.bound_to.isin(route_before_node)|
                        (restriction_condition_df.bound_to.isna()))
        terminal_mask = ((restriction_condition_df.terminal == terminal_of_call)|
                         (restriction_condition_df.terminal.isna()))
        previous_terminal_mask = ((restriction_condition_df.visited_terminal == previous_terminal_of_call)|
                                  (restriction_condition_df.visited_terminal.isna()))
        restriction_mask = node_mask&from_node_mask&to_node_mask&length_mask&draught_mask&terminal_mask&ukc_mask&terminal_mask&previous_terminal_mask
        restriction_condition_df = restriction_condition_df[restriction_mask]
        if restriction_condition_df.empty:
            return restriction, no_tidal_window

        no_tidal_window = False
        # conditions_df = restriction_condition_df[restriction_mask]
        restriction = restriction_condition_df.iloc[0]

        return restriction, no_tidal_window

    def provide_water_depth(self,vessel,node,delay=0):
        hydromanager = HydrodynamicDataManager()
        node_index = list(hydromanager.hydrodynamic_data["STATION"][:]).index(node)
        time_index = np.absolute(
            hydromanager.hydrodynamic_times - pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now + delay)).to_datetime64()
        ).argmin()
        water_level = hydromanager.hydrodynamic_data["Water level"][time_index, node_index].data
        MBL = hydromanager.hydrodynamic_data["MBL"][time_index, node_index].data
        available_water_depth = water_level - MBL
        return MBL,water_level,available_water_depth

    def provide_ukc_clearance(self,vessel,node,delay=0):
        """Function: calculates the sail-in-times for a specific vssel with certain properties and a pre-determined route and provides this information to the vessel

        Input:
            - vessel: an identity which is Identifiable, Movable, and Routable, and has VesselProperties
            - node:
            - components_calc:

        """
        MBL,water_level,available_water_depth = self.provide_water_depth(vessel,node,delay)
        ukc_s, ukc_p, ukc_r, fwa = np.zeros(4)
        ship_related_factors = {'ukc_s': ukc_s, 'ukc_p': ukc_p, 'ukc_r': ukc_r, 'fwa': fwa,'extra_ukc': vessel.metadata['ukc']}
        if 'Vertical tidal restriction' in vessel.multidigraph.nodes[node].keys():
            restrictions, _ = self.provide_tidal_window_restriction(vessel, [node], node, delay)
            if restrictions.empty:
                return [], [], available_water_depth, 0., ship_related_factors, MBL
            specifications = self.graph.nodes['A']['Vertical tidal restriction']['Specifications']
            specifications = self.graph.nodes["A"]["Vertical tidal restriction"]["Specifications"]
            ukcs_s = []
            ukcs_p = []
            ukcs_r = []
            fwas = []
            for _,specs in specifications.iterrows():
                ukcs_s.append(specs.Restriction.ukc_s)
                ukcs_p.append(specs.Restriction.ukc_p)
                ukcs_r.append(specs.Restriction.ukc_r)
                fwas.append(specs.Restriction.fwa)

            # Determine which restriction applies to vessel
            restriction_index = 0

            # Calculate ukc policy based on the applied restriction
            ukc_s = ukcs_s[restriction_index]
            ukc_p = ukcs_p[restriction_index] * vessel.T
            ukc_r = ukcs_r[restriction_index][0] * (vessel.T - ukcs_r[restriction_index][1])
            fwa = fwas[restriction_index] * vessel.T
        ship_related_factors = {"ukc_s": ukc_s, "ukc_p": ukc_p, "ukc_r": ukc_r, "fwa": fwa, "extra_ukc": vessel.metadata["ukc"]}
        required_water_depth = vessel.T + sum(ship_related_factors.values())
        net_ukc = available_water_depth - required_water_depth
        gross_ukc = available_water_depth - vessel.T
        return net_ukc, gross_ukc, available_water_depth, required_water_depth, ship_related_factors, MBL

    def provide_minimum_available_water_depth_along_route(self,vessel, route, time_start, time_end, delay=0):
        """Function: calculates the minimum available water depth (predicted/modelled/measured water level minus the local maintained bed level) along the route over time,
                      subtracted with the difference between the gross ukc and net ukc (hence: subtracted with additional safety margins consisting of vessel-related factors
                      and water level factors). The bottom-related factors are already accounted for in the use of the MBL instead of the actual depth.

        Input:
            - vessel: an identity which is Identifiable, Movable, and Routable, and has VesselProperties
            - route: a list of strings of node names that resemble the route that the vessel is planning
            to sail (can be different than vessel.route)
            - delay:
        """
        hydromanager = HydrodynamicDataManager()
        time_start_index = np.max(
            [0, np.absolute(hydromanager.hydrodynamic_times - (time_start + np.timedelta64(int(delay), "s"))).argmin() - 2]
        )
        time_end_index = np.absolute(hydromanager.hydrodynamic_times - (time_end + np.timedelta64(int(delay), "s"))).argmin()
        net_ukc = pd.DataFrame()
        times = hydromanager.hydrodynamic_times[time_start_index:time_end_index]
        start_time = times[0]
        end_time = times[-1]
        t_step = times[1] - times[0]
        time_range = np.arange(start_time, end_time + t_step, t_step)
        nodes_of_interest = route.copy()
        for node in route:
            if self.graph.nodes[node]['LAT']-self.graph.nodes[node]['MBL'] >= vessel.T+np.max([1.0,vessel.T*1.125]):
                nodes_of_interest.remove(node)

        if not nodes_of_interest:
            net_ukc['min_net_ukc'] = 0.0
            net_ukc['station'] = ''
            net_ukc.loc[time_start,:] = [1.0,route[0]]
            net_ukc.loc[time_end,:] = [1.0, route[-1]]
            return net_ukc

        # Start of calculation by looping over the nodes of the route
        for route_index, node_name in enumerate(nodes_of_interest):
            station_index = list(hydromanager.hydrodynamic_data["STATION"][:]).index(node_name)
            sailing_time_to_next_node = vessel.env.vessel_traffic_service.provide_sailing_time(vessel,route[:(route_index+1)])
            time_correction_index = int(np.round(sailing_time_to_next_node['Time'].sum() / (t_step/np.timedelta64(1, 's'))))
            water_level = hydromanager.hydrodynamic_data["Water level"][time_start_index:time_end_index, station_index].data
            MBL = hydromanager.hydrodynamic_data["MBL"][time_start_index:time_end_index, station_index].data
            _, _, _, required_water_depth, _, _ = self.provide_ukc_clearance(vessel,node_name,delay)
            water_depth = water_level - MBL
            net_ukc = pd.concat(
                [
                    net_ukc,
                    pd.DataFrame(
                        [available_water_depth - required_water_depth for available_water_depth in water_depth],
                        index=[t - time_correction_index * t_step for t in times],
                        columns=[node_name],
                    ),
                ],
                axis=1,
            )

        # Pick the minimum of the water depths for each time and each node
        net_ukc = net_ukc.dropna(axis=0)
        net_ukc["min_net_ukc"] = net_ukc.min(axis=1)
        return net_ukc

    def provide_vertical_tidal_windows(self, vessel, route, time_start, time_end, delay=0, plot=False):
        """Function: calculates the windows available to sail-in and -out of the port given the
          vertical tidal restrictions according to the tidal window policy.

        Input:
            - vessel: an identity which is Identifiable, Movable, and Routable, and has VesselProperties
            - route: a list of strings of node names that resemble the route that the vessel is planning
              to sail (can be different than vessel.route)
            - sailing_time_correction: a bool that indicates whether the calculation should correct for
              sailing_speed (dynamic calculation) or not (static calculation)

        """
        hydromanager = HydrodynamicDataManager()
        vertical_tidal_accessibility = pd.DataFrame(columns=['Tidal period', 'Period number', 'Limit', 'Accessibility'])
        time_start_index = np.max(
            [0, np.absolute(hydromanager.hydrodynamic_times - (time_start + np.timedelta64(int(delay), "s"))).argmin() - 2]
        )
        time_end_index = np.absolute(hydromanager.hydrodynamic_times - (time_end + np.timedelta64(int(delay), "s"))).argmin()

        net_ukcs = self.provide_minimum_available_water_depth_along_route(vessel, route, time_start, time_end, delay)
        new_net_ukcs = pd.DataFrame()
        for station in list(dict.fromkeys(net_ukcs['station'])):
            station_index = list(hydromanager.hydrodynamic_data["STATION"][:]).index(station)
            tidal_periods = self.read_tidal_periods(hydromanager.hydrodynamic_data, "Vertical tidal periods", station_index)
            if tidal_periods == None:
                tidal_periods = [
                    [hydromanager.hydrodynamic_times[0], "Flood start"],
                    [hydromanager.hydrodynamic_times[-1], "Flood stop"],
                ]
            tidal_periods = pd.DataFrame(tidal_periods, columns=['Period start', 'Tidal period'])
            tidal_periods = tidal_periods.reset_index(names='Period number')
            tidal_periods = tidal_periods.set_index('Period start')
            net_ukc = net_ukcs[net_ukcs.station == station]
            data = pd.concat([net_ukc, tidal_periods])
            data = data.sort_index()
            data[['Period number','Tidal period']] = data[['Period number','Tidal period']].ffill()
            data[['Period number','Tidal period']] = data[['Period number','Tidal period']].bfill()
            data = data.dropna()
            new_net_ukcs = pd.concat([new_net_ukcs,data])

        net_ukcs = new_net_ukcs.sort_index()
        net_ukcs['Period number'] = net_ukcs['Period number'].astype(int)
        net_ukcs['Tidal period'] = [tidal_period.split(' ')[0] for tidal_period in net_ukcs['Tidal period']]
        net_ukcs = net_ukcs[['Tidal period', 'Period number', 'min_net_ukc']]

        # Determine zero crossings
        zero_crossings = np.where(np.diff(np.sign(net_ukcs['min_net_ukc'])))[0]

        for iloc in zero_crossings:
            net_ukc = net_ukcs.iloc[iloc]
            if net_ukc['min_net_ukc'] > 0:
                vertical_tidal_accessibility.loc[net_ukc.name] = [net_ukc['Tidal period'], net_ukc['Period number'], 0, 'Inaccessible']
            else:
                vertical_tidal_accessibility.loc[net_ukc.name] = [net_ukc['Tidal period'], net_ukc['Period number'], 0, 'Accessible']

        # Default values
        if net_ukcs.iloc[0]['min_net_ukc'] < 0:
            vertical_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_start.replace(tzinfo=None))] = [np.nan, np.nan, 0,'Inaccessible']
        else:
            vertical_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_start.replace(tzinfo=None))] = [np.nan, np.nan, 0, 'Accessible']

        if vertical_tidal_accessibility.iloc[-1,-1] == 'Accessible':
            vertical_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_stop.replace(tzinfo=None))] = [np.nan, np.nan, 0,'Inaccessible']
        else:
            vertical_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_stop.replace(tzinfo=None))] = [np.nan, np.nan, 0,'Accessible']

        vertical_tidal_accessibility = vertical_tidal_accessibility.sort_index()
        vertical_tidal_accessibility[['Period number', 'Tidal period']] = vertical_tidal_accessibility[['Period number', 'Tidal period']].ffill()
        vertical_tidal_accessibility[['Period number', 'Tidal period']] = vertical_tidal_accessibility[['Period number', 'Tidal period']].bfill()

        # Return the sail-in or -out-times given the vertical tidal restrictions over the route of the vessel
        vertical_tidal_accessibility['Condition'] = 'Water level'
        vertical_tidal_accessibility = vertical_tidal_accessibility[~(vertical_tidal_accessibility['Accessibility'] == vertical_tidal_accessibility['Accessibility'].shift(1))]
        vertical_tidal_windows = [
            [window_start[0], window_end[0]]
            for window_start, window_end in zip(
                vertical_tidal_accessibility.iloc[:-1].iterrows(), vertical_tidal_accessibility.iloc[1:].iterrows()
            )
            if window_start[1].Accessibility == "Accessible"
        ]

        if plot:
            # Create figure
            fig, ax = plt.subplots(figsize=[16 * 2 / 3, 6])

            # Plot vertical tidal windows
            for window in vertical_tidal_windows:
                (vertical_tidal_window,) = ax.fill(
                    [window[0], window[0], window[1], window[1]],
                    [-1.5, 1.5, 1.5, -1.5],
                    facecolor="C0",
                    alpha=0.25,
                    edgecolor="none",
                )

            # Plot net UKC
            # net_ukc = self.provide_minimum_available_water_depth_along_route(vessel,route,time_start, time_end, delay)
            (net_UKC,) = ax.plot(net_ukcs["min_net_ukc"], color="C0", linewidth=2)
            ax.axhline(0, color="k", linewidth=0.5)

            # Figure bounds
            ax.set_xlim(hydromanager.hydrodynamic_times[time_start_index], hydromanager.hydrodynamic_times[time_end_index - 36])
            ax.set_ylim(-1.5, 1.5)

            # Figure ticks
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="right")
            ax.xaxis.set_major_formatter(dates.DateFormatter("%Y-%m-%d %H:%M"))

            # Figure axes
            ax.set_xlabel("Date")
            ax.set_ylabel("Net UKC [m]")

            # Legend
            ax.legend(
                [net_UKC, vertical_tidal_window],
                ["Net UKC", "Vertical tidal windows"],
                frameon=False,
                loc="upper left",
                bbox_to_anchor=(1.0, 1.0),
            )
            fig.tight_layout()

            plt.show()

        return vertical_tidal_accessibility, vertical_tidal_windows, net_ukcs

    def provide_horizontal_tidal_windows(self, vessel, route, time_start, time_end, delay=0, plot=False):
        hydromanager = HydrodynamicDataManager()
        def calculate_horizontal_tidal_window(vessel, time_start_index, time_end_index, current_velocity_station, critical_limits, restriction_type, flood=True, ebb=True,decreasing=False):
            station_index = list(hydromanager.hydrodynamic_data["STATION"][:]).index(current_velocity_station)
            time_step = hydromanager.hydrodynamic_times[1] - hydromanager.hydrodynamic_times[0]
            time_start_index = np.max([0,time_start_index-int(np.timedelta64(12,'h')/time_step)])
            currents_time = hydromanager.hydrodynamic_times[time_start_index:time_end_index]
            currents_data,_ = self.provide_governing_current_velocity(vessel,current_velocity_station,time_start_index,time_end_index)
            tidal_periods = self.read_tidal_periods(hydromanager.hydrodynamic_data, "Horizontal tidal periods", station_index)
            tidal_periods = [condition for condition in tidal_periods if condition[0] <= currents_time[-1] and condition[0] >= currents_time[0]]
            currents = pd.DataFrame({'Current velocity': currents_data}, index=currents_time)
            currents = abs(currents)
            tidal_periods = pd.DataFrame(tidal_periods,columns=['Period start','Tidal period'])
            tidal_periods = tidal_periods.reset_index(names='Period number')
            tidal_periods = tidal_periods.set_index('Period start')
            currents = pd.concat([currents,tidal_periods])
            currents = currents.sort_index()
            for column_name in ['Tidal period','Period number']:
                currents[column_name] = currents[column_name].fillna(method='ffill')
                currents[column_name] = currents[column_name].fillna(method='bfill')
            currents['Tidal period'] = [tidal_period.split(' ')[0] for tidal_period in currents['Tidal period']]
            currents['Current velocity'] = currents['Current velocity'].interpolate()
            currents["Period number"] = currents["Period number"].astype(int)

            roots_cv = [
                root for root in roots if root >= currents_time[0].astype(float) and root <= currents_time[-1].astype(float)
            ]
            times_horizontal_tidal_period = []
            for root in roots_cv:
                root = pd.Timestamp(root).to_datetime64()
                index_current_root = bisect.bisect_right(currents_time, root) - 2
                if index_current_root == -1:
                    index_current_root = index_current_root + 1
                if len(currents_data[index_prev_root:index_current_root]) == 0:
                    continue
                cvel_diff_cross = currents_data[index_current_root + 1] - currents_data[index_current_root - 1]
                if cvel_diff_cross < 0:
                    times_horizontal_tidal_period.append([root, "Ebb Start"])
                    index_prev_root = index_current_root
                elif cvel_diff_cross > 0:
                    times_horizontal_tidal_period.append([root, "Flood Start"])
                    index_prev_root = index_current_root

            tidal_periods = [condition for condition in times_horizontal_tidal_period if condition[0] <= currents_time[-1]]
            currents_time = np.append(
                currents_time, np.array([tide[0] for tide in tidal_periods if tide[0] not in currents_time], dtype="datetime64[ns]")
            )
            currents_data = [abs(value) for value in currents_data]
            currents_data = np.append(currents_data, -999 * np.ones(len(tidal_periods)))
            currents_time, currents_data = [np.array(data) for data in zip(*sorted(zip(currents_time, currents_data)))]
            # Find the intersection points with critical current velocity
            current_intersections = []
            if isinstance(cross_current_limit_dataframe, pd.DataFrame) and not cross_current_limit_dataframe.empty:
                critical_limit = np.interp(
                    currents_time.astype("float"),
                    cross_current_limit_dataframe.index.to_numpy().astype("float"),
                    cross_current_limit_dataframe.Limit.to_numpy(),
                )
                idx = np.argwhere(np.diff(np.sign(critical_limit - currents_data))).flatten()
                roots = currents_time[idx]
                current_intersections.extend(roots.astype(dtype="datetime64[ns]"))
                critical_current_velocity = np.interp(
                    np.array(current_intersections).astype("float"), currents_time.astype("float"), critical_limit
                )
                horizontal_tidal_accessibility = pd.DataFrame(
                    data=critical_current_velocity, columns=["Limit"], index=current_intersections
                )
            elif isinstance(critical_limits, list):
                critical_current_velocity = []
                for critical_limit in critical_limits:
                    idx = np.argwhere(
                        np.diff(np.sign([current_velocity - critical_limit for current_velocity in currents_data]))
                    ).flatten()
                    roots = currents_time[idx]
                    current_intersections.extend(roots.astype(dtype="datetime64[ns]"))
                    critical_current_velocity.extend(np.ones(len(roots)) * critical_limit)
                horizontal_tidal_accessibility = pd.DataFrame(
                    data=critical_current_velocity, columns=["Limit"], index=current_intersections
                )
            horizontal_tidal_accessibility = horizontal_tidal_accessibility.sort_index()

            # Determine the tidal period of the found interpolation points
            horizontal_tidal_accessibility["Period"] = ""
            horizontal_tidal_accessibility["Period_nr"] = -999
            for period_nr, (tidal_period_start, tidal_period_end) in enumerate(zip(tidal_periods[:-1], tidal_periods[1:])):
                tidal_period = tidal_period_start[1].split(" ")[0]
                if tidal_period == "Rising":
                    tidal_period = "Flood"
                if tidal_period == "Falling":
                    tidal_period = "Ebb"
                tidal_period_start = tidal_period_start[0]
                tidal_period_end = tidal_period_end[0]
                horizontal_tidal_accessibility.loc[
                    horizontal_tidal_accessibility[
                        (horizontal_tidal_accessibility.index >= tidal_period_start)
                        & (horizontal_tidal_accessibility.index <= tidal_period_end)
                    ].index,
                    "Period",
                ] = tidal_period
                horizontal_tidal_accessibility.loc[
                    horizontal_tidal_accessibility[
                        (horizontal_tidal_accessibility.index >= tidal_period_start)
                        & (horizontal_tidal_accessibility.index <= tidal_period_end)
                    ].index,
                    "Period_nr",
                ] = period_nr
            horizontal_tidal_accessibility["Period_nr"] = horizontal_tidal_accessibility["Period_nr"].astype(int)

            # Filter the found interpolation points: remove from errors
            # (multiple interpolated numbers and flood/ebb values if not required)
            if decreasing:
                selected_horizontal_tidal_accessibility = pd.DataFrame(columns=horizontal_tidal_accessibility.columns)
                if flood:
                    selected_horizontal_tidal_accessibility = pd.concat(
                        [
                            selected_horizontal_tidal_accessibility,
                            horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period == "Flood"],
                        ]
                    )
                if ebb:
                    selected_horizontal_tidal_accessibility = pd.concat(
                        [
                            selected_horizontal_tidal_accessibility,
                            horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period == "Ebb"],
                        ]
                    )
                horizontal_tidal_accessibility = selected_horizontal_tidal_accessibility.sort_index()
                horizontal_tidal_accessibility = horizontal_tidal_accessibility.loc[
                    horizontal_tidal_accessibility.index.drop_duplicates(keep=False)
                ]

            # Correct the found interpolation points
            if decreasing:
                tide_number = horizontal_tidal_accessibility.iloc[0]["Period_nr"]
                number_of_tidal_periods = horizontal_tidal_accessibility.iloc[-1]["Period_nr"]
                end_time_windows = []
                for period_nr in [
                    idx for idx, count in horizontal_tidal_accessibility.value_counts("Period_nr").items() if count == 2
                ]:
                    sub_df = horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period_nr == period_nr]
                    end_time_windows.append(tidal_periods[period_nr + 1][0] - sub_df.iloc[-1].name)
                mean_end_time_window = np.mean(end_time_windows)
                missing_tides = set(list(np.arange(tide_number, number_of_tidal_periods, 2))) - set(
                    list(dict.fromkeys(horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period == "Flood"].Period_nr))
                )
                for tide_index in missing_tides:
                    starting_time = np.datetime64(tidal_periods[tide_index][0])
                    closing_time = np.datetime64(tidal_periods[tide_index + 1][0] - mean_end_time_window)
                    next_index = bisect.bisect_right(currents_time, closing_time)
                    previous_index = next_index - 1
                    current_velocity = np.interp(
                        closing_time,
                        [currents_time[previous_index], currents_time[next_index]],
                        [currents_data[previous_index], currents_data[next_index]],
                    )
                    horizontal_tidal_accessibility.loc[starting_time, :] = [0, "Flood", tide_index]
                    horizontal_tidal_accessibility.loc[closing_time, :] = [current_velocity, "Flood", tide_index]

                for period_nr, count in [
                    (idx, count) for idx, count in horizontal_tidal_accessibility.value_counts("Period_nr").items() if count != 2
                ]:
                    if count == 1:
                        for iloc, (loc, info) in enumerate(
                            horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period_nr == period_nr].iterrows()
                        ):
                            if not iloc % 2 and info.Limit != np.max(critical_limits):
                                starting_time = np.datetime64(tidal_periods[info.Period_nr][0])
                                horizontal_tidal_accessibility.loc[starting_time] = info
                                horizontal_tidal_accessibility.loc[starting_time, "Limit"] = np.max(critical_limits)
                                break
                    else:
                        horizontal_tidal_accessibility = horizontal_tidal_accessibility.drop(
                            horizontal_tidal_accessibility[
                                (horizontal_tidal_accessibility.Period_nr == period_nr)
                                & (horizontal_tidal_accessibility.Limit == np.max(critical_limits))
                            ].index[:-1]
                        )
                        horizontal_tidal_accessibility = horizontal_tidal_accessibility.drop(
                            horizontal_tidal_accessibility[
                                (horizontal_tidal_accessibility.Period_nr == period_nr)
                                & (horizontal_tidal_accessibility.Limit == np.min(critical_limits))
                            ].index[:-1]
                        )
                        if len(horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period_nr == period_nr]) < 2:
                            for iloc, (loc, info) in enumerate(
                                horizontal_tidal_accessibility[horizontal_tidal_accessibility.Period_nr == period_nr].iterrows()
                            ):
                                if not iloc % 2 and info.Limit != np.max(critical_limits):
                                    starting_time = np.datetime64(tidal_periods[info.Period_nr][0])
                                    horizontal_tidal_accessibility.loc[starting_time] = info
                                    horizontal_tidal_accessibility.loc[starting_time, "Limit"] = np.max(critical_limits)
                                    break

            else:
                for period_nr in [
                    idx for idx, count in horizontal_tidal_accessibility.value_counts("Period_nr").items() if count % 2
                ]:
                    for loc, info in horizontal_tidal_accessibility[
                        horizontal_tidal_accessibility.Period_nr == period_nr
                    ].iterrows():
                        if (
                            currents_data[list(currents_time).index(loc) - 1] < info.Limit
                            and currents_data[list(currents_time).index(loc) + 1] < info.Limit
                        ):
                            horizontal_tidal_accessibility[loc + np.timedelta64(1, "ns")] = info
                            break

            # Add fully (in)accessible tides
            for tide in tides:
                critical_limit = critical_limits[tide]
                if critical_limit == tidal_window_constructor.accessibility.accessible.value:
                    for period_time,period_info in tidal_periods.iterrows():
                        if period_info['Tidal period'].find(tide)+1:
                            horizontal_tidal_accessibility.loc[period_time] = [critical_limit, 'Accessible', 'Current velocity', tide, period_info['Period number']]
                elif critical_limit == tidal_window_constructor.accessibility.inaccessible.value:
                    for period_time,period_info in tidal_periods.iterrows():
                        if period_info['Tidal period'].find(tide)+1:
                            horizontal_tidal_accessibility.loc[period_time] = [critical_limit, 'Inaccessible', 'Current velocity', tide, period_info['Period number']]
            last_tidal_period = tidal_periods.iloc[-1]
            horizontal_tidal_accessibility.loc[last_tidal_period.name] = [0.0, 'Inaccessible', 'Current velocity', last_tidal_period['Tidal period'], last_tidal_period['Period number']]
            return horizontal_tidal_accessibility

        # Start calculation
        horizontal_tidal_restriction_nodes = []
        horizontal_tidal_restriction_stations = []
        restrictions = pd.DataFrame()
        horizontal_tidal_accessibility = pd.DataFrame(columns=['Limit', 'Condition', 'Accessibility','Period_nr'])
        horizontal_tidal_window = False
        time_start_index = np.max([0, np.absolute(hydromanager.hydrodynamic_times - (time_start)).argmin() - 2])
        time_end_index = np.absolute(hydromanager.hydrodynamic_times - (time_end)).argmin()
        for route_index, node_name in enumerate(route):
            if 'Horizontal tidal restriction' in vessel.multidigraph.nodes[node_name].keys():
                sailing_time_to_next_node = self.provide_sailing_time(vessel, route[:(route_index + 1)])
                restriction, no_tidal_window = self.provide_tidal_window_restriction(vessel, route, node_name,sailing_time_to_next_node.Time.sum(),restriction_type='Horizontal')
                if no_tidal_window:
                    continue

                restrictions = pd.concat([restrictions,pd.DataFrame([restriction])])
                time_start_index = np.max(
                    [0, np.absolute(hydromanager.hydrodynamic_times - (time_start + np.timedelta64(int(delay), "s"))).argmin() - 2]
                )
                time_end_index = np.absolute(
                    hydromanager.hydrodynamic_times - (time_end + np.timedelta64(int(delay), "s"))
                ).argmin()

                horizontal_tidal_window = True
                current_velocity_station = restriction.Data
                cross_current_limit = {}
                cross_current_limit['Flood'] = restriction.Restriction.current_velocity_values['Flood']
                cross_current_limit['Ebb'] = restriction.Restriction.current_velocity_values['Ebb']
                if restriction.Restriction.window_method == 'Maximum':
                    next_horizontal_tidal_accessibility = calculate_horizontal_tidal_window(vessel,time_start_index,time_end_index,current_velocity_station,cross_current_limit,restriction.Restriction)
                if restriction.Restriction.window_method == 'Point-based':
                    if isinstance(restriction.Restriction.current_velocity_values['Flood'], list) and restriction.Restriction.current_velocity_values['Ebb'] == -999:
                        next_horizontal_tidal_accessibility = calculate_horizontal_tidal_window(vessel,time_start_index,time_end_index,current_velocity_station,cross_current_limit,restriction.Restriction,ebb=False,decreasing=True)
                    elif isinstance(restriction.Restriction.current_velocity_values['Ebb'], list) and restriction.Restriction.current_velocity_values['Flood'] == -999:
                        next_horizontal_tidal_accessibility = calculate_horizontal_tidal_window(vessel,time_start_index,time_end_index,current_velocity_station,cross_current_limit,restriction.Restriction,flood=False,decreasing=True)
                    else:
                        next_horizontal_tidal_accessibility = calculate_horizontal_tidal_window(vessel,time_start_index,time_end_index,current_velocity_station,cross_current_limit,restriction.Restriction,decreasing=True)

                horizontal_tidal_restriction_nodes.append(node_name)
                horizontal_tidal_restriction_stations.append(current_velocity_station)
                next_horizontal_tidal_accessibility_time_correction = np.timedelta64(int(sailing_time_to_next_node['Time'].sum()), 's')
                next_horizontal_tidal_accessibility.index -= next_horizontal_tidal_accessibility_time_correction
                next_horizontal_tidal_accessibility = next_horizontal_tidal_accessibility.sort_index()
                next_horizontal_tidal_accessibility = next_horizontal_tidal_accessibility[~(next_horizontal_tidal_accessibility['Accessibility'] == next_horizontal_tidal_accessibility['Accessibility'].shift(1))]
                if horizontal_tidal_accessibility.empty:
                    horizontal_tidal_accessibility = next_horizontal_tidal_accessibility
                else:
                    horizontal_tidal_accessibility = self.combine_tidal_windows(
                        horizontal_tidal_accessibility, next_horizontal_tidal_accessibility, current_velocity_windows=True
                    )

                horizontal_tidal_restriction_nodes.append(node_name)
                horizontal_tidal_restriction_stations.append(station)
                next_horizontal_tidal_accessibility_time_correction = np.timedelta64(
                    int(sailing_time_to_next_node["Time"].sum()), "s"
                )
                next_horizontal_tidal_accessibility.index -= next_horizontal_tidal_accessibility_time_correction
                if horizontal_tidal_accessibility.empty:
                    horizontal_tidal_accessibility = next_horizontal_tidal_accessibility
                else:
                    horizontal_tidal_accessibility = self.combine_tidal_windows(
                        horizontal_tidal_accessibility, next_horizontal_tidal_accessibility
                    )

        if horizontal_tidal_accessibility.empty or not horizontal_tidal_window:
            horizontal_tidal_accessibility = pd.DataFrame(columns=["Limit", "Condition", "Accessibility"])
            horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_start.replace(tzinfo=None)), :] = [
                0,
                "Current velocity",
                "Accessible",
            ]
            horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_stop.replace(tzinfo=None)), :] = [
                0,
                "Current velocity",
                "Inaccessible",
            ]
        else:
            if horizontal_tidal_accessibility.iloc[0].Accessibility == "Inaccessible":
                horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_start.replace(tzinfo=None)), :] = [
                    0,
                    "Current velocity",
                    "Accessible",
                ]
            else:
                horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_start.replace(tzinfo=None)), :] = [
                    0,
                    "Current velocity",
                    "Inaccessible",
                ]
            if horizontal_tidal_accessibility.iloc[-1].Accessibility == "Inaccessible":
                horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_stop.replace(tzinfo=None)), :] = [
                    0,
                    "Current velocity",
                    "Accessible",
                ]
            else:
                horizontal_tidal_accessibility.loc[np.datetime64(vessel.env.simulation_stop.replace(tzinfo=None)), :] = [
                    0,
                    "Current velocity",
                    "Inaccessible",
                ]

        horizontal_tidal_accessibility = horizontal_tidal_accessibility.sort_index()
        horizontal_tidal_windows = [
            [window_start[0], window_end[0]]
            for window_start, window_end in zip(
                horizontal_tidal_accessibility.iloc[:-1].iterrows(), horizontal_tidal_accessibility.iloc[1:].iterrows()
            )
            if window_start[1].Accessibility == "Accessible"
        ]

        if plot:
            # Create figure
            fig, ax = plt.subplots(figsize=[16 * 2 / 3, 6])

            # Plot vertical tidal windows
            for window in horizontal_tidal_windows:
                (horizontal_tidal_window,) = ax.fill(
                    [window[0], window[0], window[1], window[1]],
                    [-1.5, 1.5, 1.5, -1.5],
                    facecolor="firebrick",
                    alpha=0.25,
                    edgecolor="none",
                )

            # Plot governing current velocity
            for node,station in zip(horizontal_tidal_restriction_nodes,horizontal_tidal_restriction_stations):
                governing_current_velocity,_ = self.provide_governing_current_velocity(vessel,station,time_start_index,time_end_index)
                horizontal_tidal_accessibility_time_correction = np.timedelta64(int(self.provide_sailing_time(vessel, route[:(route.index(node) + 1)])['Time'].sum() + delay),'s')
                (current_velocity,) = ax.plot(
                    [
                        time - horizontal_tidal_accessibility_time_correction
                        for time in hydromanager.hydrodynamic_times[time_start_index:time_end_index]
                    ],
                    governing_current_velocity,
                    color="firebrick",
                    linewidth=2,
                )

            ax.axhline(0, color='k', linewidth=1)

            # Figure bounds
            ax.set_xlim(hydromanager.hydrodynamic_times[time_start_index], hydromanager.hydrodynamic_times[time_end_index - 36])
            ax.set_ylim(-1.5, 1.5)

            # Figure ticks
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="right")
            ax.xaxis.set_major_formatter(dates.DateFormatter("%Y-%m-%d %H:%M"))

            # Figure axes
            ax.set_xlabel("Date")
            ax.set_ylabel("Current velocity [m/s]")

            # Legend
            ax.legend(
                [current_velocity, horizontal_tidal_window],
                ["Current velocity", "Horizontal tidal windows"],
                frameon=False,
                loc="upper left",
                bbox_to_anchor=(1.0, 1.0),
            )
            fig.tight_layout()
            plt.show()

        return horizontal_tidal_accessibility, horizontal_tidal_windows, restrictions

    def combine_tidal_windows(self, tidal_window_1, tidal_window_2):
        accessibility_1 = tidal_window_1.iloc[
            [bisect.bisect_right(tidal_window_1.index, index) - 1 for index in tidal_window_2.index]
        ].Accessibility.to_numpy()
        accessibility_2 = tidal_window_2.iloc[
            [bisect.bisect_right(tidal_window_2.index, index) - 1 for index in tidal_window_1.index]
        ].Accessibility.to_numpy()
        accessibility_1_interpolated = tidal_window_1.copy()
        accessibility_2_interpolated = tidal_window_2.copy()
        for key, value in zip(tidal_window_2.index, accessibility_1):
            accessibility_1_interpolated.loc[key, "Accessibility"] = value
        for key, value in zip(tidal_window_1.index, accessibility_2):
            accessibility_2_interpolated.loc[key, "Accessibility"] = value
        accessibility_1_interpolated = accessibility_1_interpolated.sort_index()
        accessibility_2_interpolated = accessibility_2_interpolated.sort_index()
        tidal_accessibility = pd.concat([accessibility_1_interpolated, accessibility_2_interpolated], axis=1)
        tidal_accessibility = tidal_accessibility.sort_index()
        tidal_accessibility_limit = [
            limit_1 if not math.isnan(limit_1) else limit_2 for limit_1, limit_2 in tidal_accessibility.Limit.to_numpy()
        ]
        tidal_accessibility_condition = [
            condition_1 if isinstance(condition_1, str) else condition_2
            for condition_1, condition_2 in tidal_accessibility.Condition.to_numpy()
        ]
        tidal_accessibility_accessibility = [
            "Accessible" if accessibility_1 == accessibility_2 and accessibility_1 == "Accessible" else "Inaccessible"
            for accessibility_1, accessibility_2 in tidal_accessibility.Accessibility.to_numpy()
        ]
        tidal_accessibility = tidal_accessibility.drop(["Limit", "Condition", "Accessibility"], axis=1)
        tidal_accessibility["Limit"] = tidal_accessibility_limit
        tidal_accessibility["Condition"] = tidal_accessibility_condition
        tidal_accessibility["Accessibility"] = tidal_accessibility_accessibility
        accessible_indexes = [
            idx for idx, accessibility in enumerate((tidal_accessibility.Accessibility == "Accessible").to_numpy()) if accessibility
        ]
        inaccessible_indexes = [
            idx
            for idx, inaccessibility in enumerate((tidal_accessibility.Accessibility == "Inaccessible").to_numpy())
            if inaccessibility
        ]
        accessible_indexes = np.array(
            [indexes[-1] for indexes in np.split(accessible_indexes, np.nonzero(np.diff(accessible_indexes) != 1) + 1)], dtype=int
        )
        inaccessible_indexes = np.array(
            [indexes[0] for indexes in np.split(inaccessible_indexes, np.nonzero(np.diff(inaccessible_indexes) != 1) + 1)],
            dtype=int,
        )
        tidal_window_indexes = np.sort(np.append(accessible_indexes, inaccessible_indexes))
        tidal_accessibility = tidal_accessibility.iloc[tidal_window_indexes]
        return tidal_accessibility

    def provide_tidal_windows(self,vessel,route,time_start,time_end,ax_left=None,ax_right=None,delay=0,plot=False):
        hydromanager = HydrodynamicDataManager()
        time_start_index = np.max(
            [0, np.absolute(hydromanager.hydrodynamic_times - (time_start + np.timedelta64(int(delay), "s"))).argmin() - 2]
        )
        time_end_index = np.absolute(hydromanager.hydrodynamic_times - (time_end + np.timedelta64(int(delay), "s"))).argmin()
        vertical_tidal_accessibility,vertical_tidal_windows,net_ukcs = self.provide_vertical_tidal_windows(vessel, route, time_start, time_end, delay)
        horizontal_tidal_accessibility,horizontal_tidal_windows,horizontal_tidal_restrictions = self.provide_horizontal_tidal_windows(vessel, route, time_start, time_end, delay)
        if not horizontal_tidal_accessibility.empty:
            tidal_accessibility = self.combine_tidal_windows(vertical_tidal_accessibility,horizontal_tidal_accessibility)
        else:
            tidal_accessibility = vertical_tidal_accessibility

        tidal_windows = [[window_start[0], window_end[0]] for window_start, window_end in zip(tidal_accessibility.iloc[:-1].iterrows(), tidal_accessibility.iloc[1:].iterrows()) if window_start[1].Accessibility == 'Accessible']

        # Plot
        if plot:
            # Create figure
            if not ax_left:
                _, ax_left = plt.subplots(figsize=[16 * 2 / 3, 6])
                ax_right = ax_left.twinx()

            # Plot net UKC
            # net_ukc = self.provide_minimum_available_water_depth_along_route(vessel, route, time_start, time_end, delay)
            (net_UKC,) = ax_left.plot(net_ukcs["min_net_ukc"], color="C0", linewidth=2, zorder=1)
            minimum_required_net_ukc = ax_left.axhline(0, color="C0", linestyle="--", linewidth=2)

            # Plot governing current velocity
            horizontal_restriction_type = None
            for index,restriction_info in horizontal_tidal_restrictions.iterrows():
                horizontal_restriction_type = restriction_info.Restriction.window_method
                governing_current_velocity,_ = self.provide_governing_current_velocity(vessel,restriction_info.Data,time_start_index, time_end_index)
                horizontal_tidal_accessibility_time_correction = np.timedelta64(int(self.provide_sailing_time(vessel, route[:(route.index(restriction_info.Node) + 1)])['Time'].sum()), 's')
                (current_velocity,) = ax_right.plot(
                    [
                        time - horizontal_tidal_accessibility_time_correction
                        for time in hydromanager.hydrodynamic_times[time_start_index:time_end_index]
                    ],
                    governing_current_velocity,
                    color="firebrick",
                    linewidth=2,
                    zorder=1,
                )
                if horizontal_restriction_type == 'Maximum':
                    critical_current_velocity = ax_right.axhline(restriction_info.Restriction.current_velocity_values['Flood'], color='firebrick',linestyle='--', linewidth=2)
                    ax_right.axhline(-1 * restriction_info.Restriction.current_velocity_values['Ebb'], color='firebrick',linestyle='--', linewidth=2)
                ax_right.set_ylim(np.floor(np.min(governing_current_velocity)),np.ceil(np.max(governing_current_velocity)))

            # Figure bounds
            ax_left.set_xlim(
                hydromanager.hydrodynamic_times[time_start_index], hydromanager.hydrodynamic_times[time_end_index - 36]
            )
            ax_left.set_ylim(
                np.min([np.floor(np.min(net_ukcs["min_net_ukc"].to_numpy())), -1.0]),
                np.max([np.ceil(np.max(net_ukcs["min_net_ukc"])), 1.0]),
            )

            # Calculate vertical and horizontal tidal windows
            vertical_tidal_window_polygons = []
            for window in vertical_tidal_windows:
                vertical_tidal_window_polygons.append(
                    Polygon(
                        [
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                        ]
                    )
                )
            horizontal_tidal_window_polygons = []
            for window in horizontal_tidal_windows:
                horizontal_tidal_window_polygons.append(
                    Polygon(
                        [
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                        ]
                    )
                )
            tidal_window_polygons = []
            for window in tidal_windows:
                tidal_window_polygons.append(
                    Polygon(
                        [
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                            Point((window[0] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[1]),
                            Point((window[1] - np.datetime64("1970-01-01")) / np.timedelta64(1, "s"), ax_left.get_ylim()[0]),
                        ]
                    )
                )

            if not isinstance(horizontal_tidal_window_polygons, Polygon):
                horizontal_tidal_window_polygons = MultiPolygon(horizontal_tidal_window_polygons)
            vertical_tidal_window_polygons = MultiPolygon(vertical_tidal_window_polygons).difference(
                horizontal_tidal_window_polygons
            )
            vertical_tidal_window_polygons = vertical_tidal_window_polygons.difference(MultiPolygon(tidal_window_polygons))
            if not isinstance(vertical_tidal_window_polygons, Polygon):
                vertical_tidal_window_polygons = MultiPolygon(vertical_tidal_window_polygons)
            horizontal_tidal_window_polygons = MultiPolygon(horizontal_tidal_window_polygons).difference(
                vertical_tidal_window_polygons
            )
            horizontal_tidal_window_polygons = horizontal_tidal_window_polygons.difference(MultiPolygon(tidal_window_polygons))

            # Plot vertical tidal windows
            if not isinstance(vertical_tidal_window_polygons, Polygon):
                for polygon in vertical_tidal_window_polygons.geoms:
                    polygon_x = []
                    for timestamp in polygon.exterior.xy[0]:
                        polygon_x.append(datetime.datetime.fromtimestamp(timestamp,tz=pytz.utc))
                    polygon_y = list(polygon.exterior.xy[1])

                    (vertical_tidal_window,) = ax_left.fill(
                        polygon_x, polygon_y, facecolor="C0", alpha=0.25, edgecolor="none", zorder=0
                    )
            elif isinstance(vertical_tidal_window_polygons, Polygon):
                polygon = vertical_tidal_window_polygons
                polygon_x = []
                for timestamp in polygon.exterior.xy[0]:
                    polygon_x.append(datetime.datetime.fromtimestamp(timestamp,tz=pytz.utc))
                polygon_y = list(polygon.exterior.xy[1])
                (vertical_tidal_window,) = ax_left.fill(
                    polygon_x, polygon_y, facecolor="C0", alpha=0.25, edgecolor="none", zorder=0
                )

            # Plot horizontal tidal windows
            if not isinstance(horizontal_tidal_window_polygons, Polygon):
                for polygon in horizontal_tidal_window_polygons.geoms:
                    polygon_x = []
                    for timestamp in polygon.exterior.xy[0]:
                        polygon_x.append(datetime.datetime.fromtimestamp(timestamp,tz=pytz.utc))
                    polygon_y = list(polygon.exterior.xy[1])
                    (horizontal_tidal_window,) = ax_left.fill(
                        polygon_x, polygon_y, facecolor="firebrick", alpha=0.25, edgecolor="none", zorder=0
                    )
            elif isinstance(horizontal_tidal_window_polygons, Polygon):
                polygon = horizontal_tidal_window_polygons
                polygon_x = []
                for timestamp in polygon.exterior.xy[0]:
                    polygon_x.append(datetime.datetime.fromtimestamp(timestamp,tz=pytz.utc))
                polygon_y = list(polygon.exterior.xy[1])
                (horizontal_tidal_window,) = ax_left.fill(
                    polygon_x, polygon_y, facecolor="firebrick", alpha=0.25, edgecolor="none", zorder=0
                )

            # Plot tidal windows
            for window in tidal_windows:
                (tidal_window,) = ax_left.fill(
                    [window[0], window[0], window[1], window[1]],
                    [ax_left.get_ylim()[0], ax_left.get_ylim()[1], ax_left.get_ylim()[1], ax_left.get_ylim()[0]],
                    facecolor="limegreen",
                    alpha=0.25,
                    edgecolor="none",
                    zorder=0,
                )
            if not tidal_windows:
                tidal_window = ax_left.fill(
                    [0, 0, 0, 0],
                    [ax_left.get_ylim()[0], ax_left.get_ylim()[1], ax_left.get_ylim()[1], ax_left.get_ylim()[0]],
                    facecolor="limegreen",
                    alpha=0.25,
                    edgecolor="none",
                    zorder=0,
                )

            # Figure ticks
            ax_left.set_xticks(ax_left.get_xticks())
            ax_left.set_xticklabels(ax_left.get_xticklabels(), rotation=45, ha="right")
            ax_left.xaxis.set_major_formatter(dates.DateFormatter("%Y-%m-%d %H:%M"))

            # Figure axes
            ax_left.set_xlabel("Date")
            ax_left.set_ylabel("Net UKC [m]")
            ax_right.set_ylabel("Current velocity [m/s]")
            # Legend and title
            if horizontal_restriction_type:
                if horizontal_restriction_type == "Maximum":
                    ax_left.legend(
                        [
                            net_UKC,
                            minimum_required_net_ukc,
                            current_velocity,
                            critical_current_velocity,
                            vertical_tidal_window,
                            horizontal_tidal_window,
                            tidal_window,
                        ],
                        [
                            "Net UKC",
                            "Required net UKC",
                            "Current velocity",
                            "Vertical tidal windows",
                            "Horizontal tidal windows",
                            "Resulting tidal windows",
                        ],
                        frameon=False,
                        loc="upper left",
                        bbox_to_anchor=(1.1, 1.0),
                    )
                else:
                    ax_left.legend(
                        [
                            net_UKC,
                            minimum_required_net_ukc,
                            current_velocity,
                            vertical_tidal_window,
                            horizontal_tidal_window,
                            tidal_window,
                        ],
                        [
                            "Net UKC",
                            "Required net UKC",
                            "Current velocity",
                            "Vertical tidal windows",
                            "Horizontal tidal windows",
                            "Resulting tidal windows",
                        ],
                        frameon=False,
                        loc="upper left",
                        bbox_to_anchor=(1.1, 1.0),
                    )
            else:
                ax_left.legend(
                    [net_UKC, minimum_required_net_ukc, vertical_tidal_window, tidal_window],
                    ["Net UKC", "Required net UKC", "Vertical tidal windows", "Resulting tidal windows"],
                    frameon=False,
                    loc="upper left",
                    bbox_to_anchor=(1.05, 1.0),
                )
            ax_left.set_title(
                f"Accessibility of {vessel.type}-class vessel with name {vessel.name} "
                f"and {np.round(vessel.T, 2)}m draught and a length of {np.round(vessel.L)}m, sailing {vessel.bound}"
            )
            plt.show()
            return tidal_accessibility, tidal_windows, ax_left, ax_right

        return tidal_accessibility, tidal_windows



from openclsim.core import Identifiable, Locatable, Log, SimpyObject

from .container import HasContainer
from .misc import ExtraMetadata, Neighbours
from .movable import ContainerDependentMovable, Movable, Routable, Routeable
from .resource import HasResource
from .vessel_properties import HasLength, HasLoad, VesselProperties
from . import logutils, plotutils

__all__ = [
    "Identifiable",
    "Locatable",
    "Log",
    "SimpyObject",
    "HasContainer",
    "ExtraMetadata",
    "Neighbours",
    "ContainerDependentMovable",
    "Movable",
    "Routable",
    "Routeable",
    "HasResource",
    "HasLength",
    "HasLoad",
    "VesselProperties",
    "logutils",
    "plotutils",
]


"""
Mixin classes related to containers.

The following classes are provided:
- HasContainer

"""
# package(s) for documentation, debugging, saving and loading
import logging

# package(s) related to the simulation
import simpy

# use OpenCLSim objects for core objects (identifiable is imported for later use)
from openclsim.core import SimpyObject

# get logger
logger = logging.getLogger(__name__)


class HasContainer(SimpyObject):
    """Mixin class: Something with a container, modelled as a storage capacity

    Parameters
    -----------
    capacity: float
        the capacity of the container, which may either be continuous (like water) or discrete (like apples)
    level: int, default=0
        level of the container at the beginning of the simulation
    total_requested: int, default=0
        total amount that has been requested at the beginning of the simulation
    args, kwargs:
        passed to SimpyObject. Must at least contain parameter env: simpy.Environment.

    Attributes
    -----------
    container: simpy.Container
        the container that is used to limit the amount that can be requested.
    total_requested: int
        total amount that has been requested.
    """

    def __init__(self, capacity: float, level: float = 0, total_requested: float = 0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """Initialization"""
        self.container = simpy.Container(self.env, capacity, init=level)
        self.total_requested = total_requested

    @property
    def is_loaded(self):
        """Return if the container is loaded"""
        return True if self.container.level > 0 else False

    @property
    def filling_degree(self):
        """return the filling degree of the container"""
        return self.container.level / self.container.capacity

    @property
    def max_load(self):
        """return the maximum cargo to load"""
        # independent of trip
        return self.container.capacity - self.container.level